(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory CPF_Parser 
imports
  Proof_Checker
  Xml
begin

fun var :: "('a, string) term xmlt" where "var x = (Xml.text ''var'' x \<guillemotright>= (return \<circ> Var))"
hide_const (open) var

partial_function (parse_result)
  "term" :: "'a xmlt \<Rightarrow> ('a, string) term xmlt"
where
  [code]: "term xml2name x = xml_options [
    (''var'', CPF_Parser.var),
    (''funapp'', xml1many_elements ''funapp'' xml2name (
      xml1element ''arg'' (term xml2name) id) Fun)
  ] x"  
hide_const (open) "term"

partial_function (parse_result) ctxt :: "'a xmlt \<Rightarrow> ('a, string) ctxt xmlt" where
  [code]: "ctxt xml2name x = xml_options [
     (''box'', xml0elements ''box'' Hole),  
     (''funContext'', xml4elements ''funContext'' 
       xml2name 
       (xml_many_elements ''before'' (CPF_Parser.term xml2name) id)
       (ctxt xml2name)
       (xml_many_elements ''after'' (CPF_Parser.term xml2name) id)
       More)
   ] x"  

definition flat_contexts :: "'a xmlt \<Rightarrow> ('a, string) ctxt list xmlt" where
  "flat_contexts xml2name = xml_many_elements ''flatContexts'' (CPF_Parser.ctxt xml2name) id"
hide_const (open) flat_contexts

definition subst :: "'a xmlt \<Rightarrow> ('a, string) substL xmlt" where
  "subst xml2name = xml_many_elements ''substitution''
    (xml2elements ''substEntry'' 
      (Xml.text ''var'')
      (CPF_Parser.term xml2name)
      Pair
    ) id"
hide_const (open) subst

primrec list2position :: "nat list \<Rightarrow> pos" where
  "list2position [] = \<epsilon>"
| "list2position (n # ns) = n <# list2position ns"

definition position :: "nat xmlt" where
  "position \<equiv> xml_change (Xml.nat ''position'') (\<lambda>n. n - 1)"
hide_const (open) position

definition pos :: "pos xmlt" where
  "pos \<equiv> xml_many_elements ''positionInTerm'' CPF_Parser.position list2position"
hide_const (open) pos


definition afs :: "'a xmlt \<Rightarrow> 'a afs_list xmlt" where
  "afs xml2name \<equiv> xml_many_elements ''argumentFilter'' (xml3elements ''argumentFilterEntry''
    xml2name
    (Xml.nat ''arity'')
    (xml_options [
      (''collapsing'', xml_change (Xml.nat ''collapsing'') (\<lambda>i. Collapse (i - 1))),
      (''nonCollapsing'', xml_many_elements ''nonCollapsing'' CPF_Parser.position AFList)
    ])
    (\<lambda>f a e. ((f, a), e))
  ) id"
hide_const (open) afs

definition proj :: "'a xmlt \<Rightarrow> 'a projL xmlt" where
  "proj xml2name \<equiv> xml_change (CPF_Parser.afs xml2name) (\<lambda>afl. Projection 
    (map (\<lambda>(fa, e). case e of 
      Collapse i \<Rightarrow> Pair fa i
    | _ \<Rightarrow> Pair fa (snd fa)) afl))" 

definition rule :: "'a xmlt \<Rightarrow> ('a, string) rule xmlt" where
  "rule xml2name = xml2elements ''rule''
    (xml1element ''lhs'' (CPF_Parser.term xml2name) id)
    (xml1element ''rhs'' (CPF_Parser.term xml2name) id)
    Pair"
hide_const (open) rule 

definition
  relstep :: "'a xmlt \<Rightarrow> (pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt"
where
  "relstep xml2name = xml3to4elements ''rewriteStep''
    CPF_Parser.pos
    (CPF_Parser.rule xml2name)
    (xml0elements ''relative'' ())
    (CPF_Parser.term xml2name)
    (\<lambda>p r rel t. (p, r, rel = None, t))"
hide_const (open) relstep

definition
  rstep :: "'a xmlt \<Rightarrow> (pos \<times> ('a, string) rule \<times> ('a, string) term) xmlt"
where
  "rstep xml2name = xml3elements ''rewriteStep''
    CPF_Parser.pos
    (CPF_Parser.rule xml2name)
    (CPF_Parser.term xml2name)
    (\<lambda>p r t. (p, r, t))"
hide_const (open) rstep

definition
  estep :: "'a xmlt \<Rightarrow> (pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt"
where
  "estep xml2name = xml4elements ''equationStep''
    CPF_Parser.pos
    (CPF_Parser.rule xml2name)
    (xml_options [
       (''leftRight'', xml0elements ''leftRight'' True),
       (''rightLeft'', xml0elements ''rightLeft'' False)
    ])
    (CPF_Parser.term xml2name)
    (\<lambda>p r b t. (p, r, b, t))"
hide_const (open) estep

definition
  relsteps :: "'a xmlt \<Rightarrow> (('a,string)term \<times> ('a,string)prseq) xmlt"
where
  "relsteps xml2name \<equiv> xml1many_elements ''rewriteSequence'' 
    (xml1element ''startTerm'' (CPF_Parser.term xml2name) id)
    (CPF_Parser.relstep xml2name)
    Pair"
hide_const (open) relsteps

definition
  rsteps :: "'a xmlt \<Rightarrow> (('a,string)term \<times> ('a,string)rseq) xmlt"
where
  "rsteps xml2name \<equiv> xml1many_elements ''rewriteSequence'' 
    (xml1element ''startTerm'' (CPF_Parser.term xml2name) id)
    (CPF_Parser.rstep xml2name)
    Pair"
hide_const (open) rsteps


definition pat_eqv_prf :: "'a xmlt \<Rightarrow> ('a,string)pat_eqv_prf xmlt"
  where "pat_eqv_prf xml2name \<equiv> let sub = CPF_Parser.subst xml2name in
   xml1element ''patternEquivalence''
    (xml_options [
       (''domainRenaming'', xml1element ''domainRenaming'' sub Pat_Dom_Renaming),
       (''irrelevant'', xml2elements ''irrelevant''  sub sub Pat_Irrelevant),
       (''simplification'', xml2elements ''simplification'' sub sub Pat_Simplify)
     ]) id"
hide_const (open) pat_eqv_prf

definition pat_term :: "'f xmlt \<Rightarrow> (('f,string)term \<times> ('f,string)substL \<times> ('f,string)substL) xmlt"
  where "pat_term xml2name \<equiv> xml3elements ''patternTerm''
            (CPF_Parser.term xml2name)
            (CPF_Parser.subst xml2name)
            (CPF_Parser.subst xml2name)
            (\<lambda> t s1 s2. (t,s1,s2))"
            

abbreviation (input) pat_rule_fun where "pat_rule_fun xml2name x pat \<equiv> (let 
   sub = CPF_Parser.subst xml2name;
   pt = pat_term xml2name;
   var = Xml.text ''var''
 in 
   xml3elements ''patternRule'' pt pt
    (xml_options [
       (''originalRule'', xml2elements ''originalRule'' (CPF_Parser.rule xml2name) (Xml.bool ''isPair'')  Pat_OrigRule),
       (''initialPumping'', xml3elements ''initialPumping'' pat sub sub Pat_InitPump),
       (''initialPumpingContext'', xml4elements ''initialPumpingContext'' pat sub CPF_Parser.pos var Pat_InitPumpCtxt),
       (''equivalence'', xml3elements ''equivalence'' pat (xml_options [
                  (''left'',xml0elements ''left'' True),
                  (''right'',xml0elements ''right'' False)
             ]) (CPF_Parser.pat_eqv_prf xml2name) Pat_Equiv),
       (''narrowing'', xml3elements ''narrowing'' pat pat CPF_Parser.pos Pat_Narrow),
       (''instantiation'', xml3elements ''instantiation'' pat sub (xml_options [
                  (''base'',xml0elements ''base'' Pat_Base),
                  (''pumping'',xml0elements ''pumping'' Pat_Pump),
                  (''closing'',xml0elements ''closing'' Pat_Close)
             ]) Pat_Inst),
       (''rewriting'', xml3elements ''rewriting'' pat (CPF_Parser.rsteps xml2name)  (xml_options [
                  (''base'',xml0elements ''base'' (Pat_Base,[])),
                  (''pumping'',xml1element ''pumping'' var (Pair Pat_Pump)),
                  (''closing'',xml1element ''closing'' var (Pair Pat_Close))
             ]) (\<lambda> p seq (po,va). Pat_Rewr p seq po va))
     ]) (\<lambda> _ _ z. z) x)" 

partial_function (parse_result) pat_rule_prf :: "'a xmlt \<Rightarrow> ('a,string)pat_rule_prf xmlt"
  where "pat_rule_prf xml2name x = pat_rule_fun xml2name x (pat_rule_prf xml2name)"

(* one cannot put the 'let pat = ... ' directly into partial function, since then the monotonicity fails *)
lemma pat_rule_prf [code]: 
  "pat_rule_prf xml2name x = (let pat = pat_rule_prf xml2name in pat_rule_fun xml2name x pat)"
     unfolding pat_rule_prf.simps[of _ x] Let_def by simp

hide_const (open) pat_rule_prf


definition
  conversion :: "'a xmlt \<Rightarrow> (('a,string)eseq) xmlt"
where
  "conversion xml2name \<equiv> xml1many_elements ''conversion'' 
    (xml1element ''startTerm'' (CPF_Parser.term xml2name) id)
    (CPF_Parser.estep xml2name)
    (\<lambda> _ x. x)"
hide_const (open) conversion

definition subsumption_proof :: "'a xmlt \<Rightarrow> ('a,string)subsumption_proof xmlt"
  where "subsumption_proof xml2name \<equiv> xml_many_elements ''subsumptionProof''
     (xml2elements ''ruleSubsumptionProof''
        (CPF_Parser.rule xml2name)
        (CPF_Parser.conversion xml2name)
        Pair
    ) id"
hide_const (open) subsumption_proof

definition joinable_critical_pairs :: "'a xmlt \<Rightarrow> ('a,string)crit_pair_joins xmlt"
  where "joinable_critical_pairs xml2name \<equiv> let rew = CPF_Parser.rsteps xml2name in
    xml_many_elements ''joinableCriticalPairs'' 
      (xml2elements ''joinableCriticalPair'' rew rew (\<lambda> (s,sseq) (t,tseq). (s,sseq,t,tseq))) id"

definition wcr_proof :: "'a xmlt \<Rightarrow> 'a join_info xmlt"
  where "wcr_proof xml2name \<equiv> xml1element ''wcrProof'' 
    (xml_options [
       (''joinableCriticalPairs'', xml_change (joinable_critical_pairs xml2name) Guided),
       (''joinableCriticalPairsBFS'', xml_change (Xml.nat ''joinableCriticalPairsBFS'') Join_BFS),
       (''joinableCriticalPairsAuto'', xml0elements ''joinableCriticalPairsAuto'' Join_NF)
  ]) id"
         
definition 
  loop ::
    "'a xmlt \<Rightarrow>
    (('a, string) term \<times> ('a, string) prseq \<times> ('a, string) substL \<times> ('a, string) ctxt) xmlt"
where
  "loop xml2name \<equiv> xml3elements ''loop''
    (CPF_Parser.relsteps xml2name)
    (CPF_Parser.subst xml2name)
    (CPF_Parser.ctxt xml2name)
    (\<lambda>(s, rseq) \<sigma> c. (s, rseq, \<sigma>, c))"
hide_const (open) loop

definition
  start_term :: "'a xmlt \<Rightarrow> ('a, string) term \<Rightarrow> ('a, string) term xmlt"
where
  "start_term xml2name t = xml1element ''startTerm'' (\<lambda>xml. (
     CPF_Parser.term xml2name xml \<guillemotright>= (\<lambda>s. if s = t
       then return t
       else error (''<startTerm> does not match lhs'')
     )
   )) id"
hide_const (open) start_term

definition
  rseq ::
    "'a xmlt \<Rightarrow> 'a proj \<Rightarrow> ('a, string) rule \<Rightarrow> (('a,string)rule \<times> ('a,string)rseq)xmlt"
where
  "rseq xml2name \<pi> r = xml1many_elements ''rewriteSequence''
     (CPF_Parser.start_term xml2name (proj_term \<pi> (fst r)))
     (CPF_Parser.rstep xml2name)
     (\<lambda>_ rseq. (r, rseq))"
hide_const (open) rseq

fun
  projected_rseq ::
    "'a xmlt \<Rightarrow> 'a proj \<Rightarrow>
    (('a, string) rule \<times> ('a,string)rseq) xmlt"
where
  "projected_rseq xml2name \<pi> (XML name atts cs text) = (
     let tag = ''projectedRewriteSequence'' in (
     if name = tag \<and> atts = [] \<and> text = None \<and> length cs = 2
       then (CPF_Parser.rule xml2name (cs ! 0) \<guillemotright>= (\<lambda>r. CPF_Parser.rseq xml2name \<pi> r (cs ! 1)))
       else no_transform tag (XML name atts cs text)))"
hide_const (open) projected_rseq

definition rules :: "'a xmlt \<Rightarrow> ('a, string) rules xmlt" where
  "rules xml2name = xml_many_elements ''rules'' (CPF_Parser.rule xml2name) id"
hide_const (open) rules


partial_function (parse_result)
  arith_fun :: "arithFun xmlt"
where
  [code]: "arith_fun xml = xml1element ''arithFunction'' (xml_options [ 
     (''natural'', xml_change (Xml.nat ''natural'') Const),
     (''variable'', xml_change (Xml.nat ''variable'') (\<lambda>n::nat. Arg (n - 1))),
     (''sum'', xml_many_elements ''sum'' arith_fun Sum),
     (''product'', xml_many_elements ''product'' arith_fun Prod),
     (''min'', xml_many_elements ''min'' arith_fun Min),
     (''max'', xml_many_elements ''max'' arith_fun Max)
   ]) id  xml"

definition
  sl_variant :: "'f xmlt \<Rightarrow> ('f,string) sl_variant xmlt"
where
  "sl_variant xml2name \<equiv> xml1element ''model'' (xml_options [
     (''finiteModel'', xml1or2many_elements ''finiteModel''
         (xml_change (Xml.nat ''carrierSize'') (\<lambda> n :: nat. n - 1))
         (xml1element ''tupleOrder'' (xml0elements ''pointWise'' ()) id)
         (xml3elements ''interpret''
           xml2name
           (Xml.nat ''arity'')
           CPF_Parser.arith_fun
           (\<lambda>f a inter. ((f, a), inter)))
         (\<lambda>n model xs. if model = None then Finitelab (SL_Inter n xs) else QuasiFinitelab (SL_Inter n xs) [])),
     (''rootLabeling'',xml_choice ''rootLabeling'' [
            xml0elements ''rootLabeling'' (Rootlab None),
            xml1element ''rootLabeling'' xml2name (\<lambda> f. Rootlab (Some (f,1)))
        ])
   ]) id"
hide_const (open) sl_variant


fun sum_lpoly :: "('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "sum_lpoly _ [] ys = ys"
| "sum_lpoly _ xs [] = xs"
| "sum_lpoly pl (x # xs) (y # ys) = pl x y # sum_lpoly pl xs ys"

fun
  lpoly_of ::
    "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow>
    (nat, 'a) tpoly \<Rightarrow> ('a \<times> 'a list) sparse_result"
where
  "lpoly_of _ _ _ _ (PNum i) = return (i, [])"
| "lpoly_of ze on pl ti (PVar x) = return (ze, replicate x ze @ [on])"
| "lpoly_of ze on pl ti (PSum []) = return (ze, [])"
| "lpoly_of ze on pl ti (PSum (p # ps)) = (
     lpoly_of ze on pl ti p 
  \<guillemotright>= (\<lambda>(cp, ncp). lpoly_of ze on pl ti (PSum ps)
  \<guillemotright>= (\<lambda>(cq, ncq). return (pl cp cq, sum_lpoly pl ncp ncq))))"
| "lpoly_of ze on pl ti (PMult []) = return (on, [])"
| "lpoly_of ze on pl ti (PMult (p # ps)) = (
     lpoly_of ze on pl ti p
  \<guillemotright>= (\<lambda>(cp, ncp). lpoly_of ze on pl ti (PMult ps)
  \<guillemotright>= (\<lambda>(cq, ncq). if (Ball (set ncp) (op = ze))
       then return (ti cp cq, map (ti cp) ncq)
       else (if (Ball (set ncq) (op = ze))
         then return (ti cp cq, map (\<lambda> x. ti x cq) ncp)
         else error ''cannot transform non-linear polynomial to linear polynomial''))))"


partial_function (parse_result) polynomial :: "'a xmlt \<Rightarrow> (nat, 'a) tpoly xmlt" where
  [code]: "polynomial xml2coeff x = xml1element ''polynomial'' (xml_options [
    (''coefficient'', xml1element ''coefficient'' xml2coeff PNum), 
    (''variable'', xml_change (Xml.nat ''variable'') (\<lambda>n. PVar (n - 1))),
    (''sum'', xml_many_elements ''sum'' (polynomial xml2coeff) PSum),
    (''product'', xml_many_elements ''product'' (polynomial xml2coeff) PMult)
  ]) id x"

definition int_coeff :: "int xmlt" where
  "int_coeff \<equiv> Xml.int ''integer''"
hide_const (open) int_coeff

definition vec_coeff :: "'a xmlt \<Rightarrow> 'a vec xmlt" where
  "vec_coeff xml2coeff \<equiv> xml_many_elements ''vector''
     (xml1element ''coefficient'' xml2coeff id) id"
hide_const (open) vec_coeff

(* allow vectors and convert them to matrices by filling in zeros *)
definition mat_coeff :: "'a \<Rightarrow> 'a xmlt \<Rightarrow> 'a mat xmlt" where
  "mat_coeff ze xml2coeff \<equiv> xml_options [
     (''matrix'', xml_many_elements ''matrix'' (CPF_Parser.vec_coeff xml2coeff) id),
     (''vector'', xml_change
       (CPF_Parser.vec_coeff xml2coeff)
       (\<lambda>v. v # replicate (length v - 1) (replicate (length v) ze))
     )]"
hide_const (open) mat_coeff

definition arctic_coeff :: "arctic xmlt" where
  "arctic_coeff \<equiv> xml_options [
    (''integer'', xml_change (Xml.int ''integer'') Num_arc),
    (''minusInfinity'', xml0elements ''minusInfinity'' MinInfty)
  ]"
hide_const (open) arctic_coeff

definition arctic_rat_coeff :: "arctic_rat xmlt" where
  "arctic_rat_coeff \<equiv> xml_options [
    (''integer'', xml_change Xml.rat Num_arc_rat),
    (''rational'', xml_change Xml.rat Num_arc_rat),
    (''minusInfinity'', xml0elements ''minusInfinity'' MinInfty_rat)
  ]"
hide_const (open) arctic_rat_coeff

datatype Domain =
  Natural nat | Arctic | Arctic_rat | Int_mat nat nat | Arc_mat nat
| Arc_rat_mat nat | Rational rat nat | Rat_mat nat nat

definition rat_domain :: "rat xmlt" where
  "rat_domain \<equiv> xml1element ''rationals'' (xml1element ''delta'' Xml.rat id) id"
hide_const (open) rat_domain

definition basic_domain :: "(nat \<Rightarrow> Domain) xmlt" where
  "basic_domain \<equiv> xml_options [
    (''naturals'', xml0elements ''naturals'' Natural),
    (''arctic'', xml1element ''arctic'' 
      (xml1element ''domain'' (xml_options [
        (''naturals'', xml0elements ''naturals'' (\<lambda> d. Arctic)),
        (''integers'', xml0elements ''integers'' (\<lambda> d. Arctic)),
        (''rationals'', xml_change CPF_Parser.rat_domain (\<lambda> r d. Arctic_rat))
      ]) id) id),
    (''rationals'', xml_change CPF_Parser.rat_domain Rational)
  ]"
hide_const (open) basic_domain

definition interpretation_type :: "Domain xmlt" where
  "interpretation_type \<equiv> (xml1element ''type'' 
     (xml_options [
       (''polynomial'', (xml2elements ''polynomial'' 
       (xml1element ''domain'' 
       (xml_options [
         (''naturals'', CPF_Parser.basic_domain),
         (''arctic'', CPF_Parser.basic_domain),
         (''rationals'', CPF_Parser.basic_domain),
         (''matrices'',xml3elements ''matrices''
           (Xml.nat ''dimension'') 
           (Xml.nat ''strictDimension'') 
           (xml1element ''domain'' CPF_Parser.basic_domain id)
           (\<lambda>di sd domain d. case domain d of
             Natural _ \<Rightarrow> Int_mat di sd
           | Rational _ _ \<Rightarrow> Rat_mat di sd
           | Arctic \<Rightarrow> Arc_mat di
           | Arctic_rat \<Rightarrow> Arc_rat_mat di
         ))
       ]) 
     id)
     (Xml.nat ''degree'') 
     (\<lambda>type d. type d))),
     (''matrixInterpretation'',xml3elements ''matrixInterpretation''
       (xml1element ''domain'' CPF_Parser.basic_domain id)
       (Xml.nat ''dimension'') 
       (Xml.nat ''strictDimension'') 
       (\<lambda>domain di sd. case domain 0 of
         Natural _ \<Rightarrow> Int_mat di sd
       | Rational _ _  \<Rightarrow> Rat_mat di sd
       | Arctic \<Rightarrow> Arc_mat di
       | Arctic_rat \<Rightarrow> Arc_rat_mat di
     ))
   ])
 id)"
hide_const (open) interpretation_type

datatype ('a, 'b, 'c, 'd, 'e, 'f, 'g, 'h, 'i, 'j) alter = Alter1 'a | Alter2 'b
| Alter3 'c | Alter4 'd | Alter5 'e | Alter6 'f | Alter7 'g | Alter8 'h
| Alter9 'i | Alter10 'j

fun alter1 where "alter1 (Alter1 x) = x"
fun alter2 where "alter2 (Alter2 x) = x"
fun alter3 where "alter3 (Alter3 x) = x"
fun alter4 where "alter4 (Alter4 x) = x"
fun alter5 where "alter5 (Alter5 x) = x"
fun alter6 where "alter6 (Alter6 x) = x"
fun alter7 where "alter7 (Alter7 x) = x"
fun alter8 where "alter8 (Alter8 x) = x"
fun alter9 where "alter9 (Alter9 x) = x"
fun alter10 where "alter10 (Alter10 x) = x"

fun fit_length :: "'a \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "fit_length _ 0 _ = []"
| "fit_length a n [] = replicate n a"
| "fit_length a (Suc n) (b # as) = b # (fit_length a n as)"

definition "interpretation" :: "'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "interpretation xml2name \<equiv> 
      (xml1many_elements_gen ''interpretation'' 
       CPF_Parser.interpretation_type
       (\<lambda>type. case type of
          Natural deg \<Rightarrow> if deg \<le> 1 then (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<guillemotright>= lpoly_of (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c))
                        (\<lambda>f a poly.  (Alter1 ((f,a),(fst poly, fit_length (poly_c.zero c) a (snd poly)))))))
                   (int_components :: (int)poly_c) else (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Alter9 ((f,a),poly))))
        | Rational d deg \<Rightarrow> if deg \<le> 1 then (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial Xml.rat x \<guillemotright>= lpoly_of (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c))
                        (\<lambda>f a poly.  (Alter2 ((f,a),(fst poly, fit_length (poly_c.zero c) a (snd poly))))))) 
                    (rat_components :: (rat)poly_c) else (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial Xml.rat x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Alter10 ((f,a),poly))))
        | Arctic \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.arctic_coeff x \<guillemotright>= lpoly_of (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c))
                        (\<lambda>f a poly.  (Alter3 ((f,a),(fst poly, fit_length (poly_c.zero c) a (snd poly))))))) 
                    (arctic_components :: (arctic)poly_c)
        | Arctic_rat \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.arctic_rat_coeff x \<guillemotright>= lpoly_of (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c))
                        (\<lambda>f a poly.  (Alter7 ((f,a),(fst poly, fit_length (poly_c.zero c) a (snd poly))))))) 
                    (arctic_rat_components :: (arctic_rat)poly_c)
        | Int_mat n sd \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff 0 CPF_Parser.int_coeff) x \<guillemotright>= lpoly_of (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c))
                        (\<lambda>f a poly.  (Alter4 ((f,a),(fst poly, fit_length (poly_c.zero c) a (snd poly))))))) 
                    (mat_components n sd (int_components :: (int)poly_c))
        | Rat_mat n sd \<Rightarrow> (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff 0 Xml.rat) x \<guillemotright>= lpoly_of (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c))
                        (\<lambda>f a poly.  (Alter5 ((f,a),(fst poly, fit_length (poly_c.zero c) a (snd poly))))))) 
                    (mat_components n sd (rat_components :: (rat)poly_c))
        | Arc_mat n \<Rightarrow> (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff MinInfty CPF_Parser.arctic_coeff) x \<guillemotright>= lpoly_of (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c))
                        (\<lambda> f a poly. (Alter6 ((f,a),(fst poly, fit_length (poly_c.zero c) a (snd poly))))))) 
                    (mat_arctic_components n (arctic_components :: (arctic)poly_c))
        | Arc_rat_mat n \<Rightarrow> (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff MinInfty_rat CPF_Parser.arctic_rat_coeff) x \<guillemotright>= lpoly_of (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c))
                        (\<lambda>f a poly. (Alter8 ((f,a),(fst poly, fit_length (poly_c.zero c) a (snd poly))))))) 
                    (mat_arctic_components n (arctic_rat_components :: (arctic_rat)poly_c))
        )
      (\<lambda> type pi. case type of 
            Natural deg \<Rightarrow> if (deg \<le> 1) then Int_carrier (map alter1 pi) else Int_nl_carrier (map alter9 pi)
          | Rational d deg \<Rightarrow> if (deg \<le> 1) then Rat_carrier (map alter2 pi) else Rat_nl_carrier d (map alter10 pi)
          | Arctic \<Rightarrow> Arctic_carrier (map alter3 pi)
          | Arctic_rat \<Rightarrow> Arctic_rat_carrier (map alter7 pi)
          | Int_mat n sd \<Rightarrow> Int_mat_carrier n sd (map alter4 pi)
          | Rat_mat n sd \<Rightarrow> Rat_mat_carrier n sd (map alter5 pi)
          | Arc_mat n \<Rightarrow> Arc_mat_carrier n (map alter6 pi)
          | Arc_rat_mat n \<Rightarrow> Arc_rat_mat_carrier n (map alter8 pi)
      ))
    "
hide_const (open) "interpretation"

definition status_precedence :: "'a xmlt \<Rightarrow> ('a status_prec_repr) xmlt" where
  "status_precedence xml2name \<equiv> xml_many_elements ''statusPrecedence'' 
  (xml4elements ''statusPrecedenceEntry''
      xml2name
      (Xml.nat ''arity'')
      (Xml.nat ''precedence'')
      (xml_options [(''lex'', xml0elements ''lex'' Lex), (''mul'', xml0elements ''mul'' Mul)])
      (\<lambda>f a p s. (((f,a), (p,s))))
  ) id" 
hide_const (open) status_precedence

definition path_order :: "'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "path_order xml2name \<equiv> xml1to2elements ''pathOrder''
   (CPF_Parser.status_precedence xml2name)
   (CPF_Parser.afs xml2name)
   (\<lambda>prec\<tau> afo. let afs = (case afo of None \<Rightarrow> [] | Some af \<Rightarrow> af)
 in
              RPO prec\<tau> afs)"
hide_const (open) path_order

definition precedence_weight :: "'a xmlt \<Rightarrow> (nat \<Rightarrow> 'a prec_weight_repr) xmlt" where
  "precedence_weight xml2name \<equiv> xml_many_elements ''precedenceWeight'' 
  (xml4elements ''precedenceWeightEntry''
      xml2name
      (Xml.nat ''arity'')
      (Xml.nat ''precedence'')
      (Xml.nat ''weight'')
      (\<lambda>f a p w. (((f,a), (p,w))))
  ) Pair" 
hide_const (open) precedence_weight

definition knuth_bendix_order :: "'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "knuth_bendix_order xml2name \<equiv> xml2to3elements ''knuthBendixOrder''
   (Xml.nat ''w0'')
   (CPF_Parser.precedence_weight xml2name)
   (CPF_Parser.afs xml2name)
   (\<lambda> w0 prw afo. let afs = (case afo of None \<Rightarrow> [] | Some af \<Rightarrow> af)
 in
              KBO (prw w0) afs)"
hide_const (open) knuth_bendix_order

definition redtriple :: "'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "redtriple xml2name = 
     xml1element ''redPair'' 
     (xml_options [
  (''interpretation'', CPF_Parser.interpretation xml2name), 
  (''pathOrder'', CPF_Parser.path_order xml2name),
  (''knuthBendixOrder'', CPF_Parser.knuth_bendix_order xml2name)
  ]) id"
hide_const (open) redtriple

definition level_mapping :: "'a xmlt \<Rightarrow> 'a scnp_af xmlt" where
  "level_mapping xml2name = xml_many_elements ''levelMapping'' 
     (xml2many_elements ''levelMappingEntry''
      xml2name
      (Xml.nat ''arity'')
      (xml2elements ''positionLevelEntry''
         (Xml.nat ''position'')
         (Xml.nat ''level'')
         Pair
      )
      (\<lambda>f a ps. ((f,a), map (\<lambda> (p,l). (if p = 0 then a else p - 1,l)) ps))
  ) id" 

definition root_redtriple :: "'a xmlt \<Rightarrow> 'a root_redtriple_impl xmlt" where
  "root_redtriple xml2name = xml1element ''redPair'' (       
    xml3elements ''scnp'' 
      (xml1element ''status'' 
         (xml_options [
             (''ms'',xml0elements ''ms'' MS_Ext),
             (''min'',xml0elements ''min'' Min_Ext),
             (''dms'',xml0elements ''dms'' Dms_Ext),
             (''max'',xml0elements ''max'' Max_Ext)
                      ])
         id)     
      (level_mapping xml2name)
      (CPF_Parser.redtriple xml2name) 
      SCNP) id"
hide_const (open) root_redtriple
      

definition
  ordering_constraint_proof :: "'a xmlt \<Rightarrow> 'a redtriple_impl xmlt"
where
  "ordering_constraint_proof xml2name = 
     xml1element ''orderingConstraintProof'' (CPF_Parser.redtriple xml2name) id"

definition
  root_ordering_constraint_proof :: "'a xmlt \<Rightarrow> 'a root_redtriple_impl xmlt"
where
  "root_ordering_constraint_proof xml2name = 
     xml1element ''orderingConstraintProof'' (CPF_Parser.root_redtriple xml2name) id"

definition
  ordering_constraint_proof' :: "'a xmlt \<Rightarrow> ('a root_redtriple_impl + 'a redtriple_impl) xmlt"
where
  "ordering_constraint_proof' xml2name = xml_choice ''reduction pair''
     [xml_change (ordering_constraint_proof xml2name) Inr,
      xml_change (root_ordering_constraint_proof xml2name) Inl]"

hide_const (open) ordering_constraint_proof
hide_const (open) root_ordering_constraint_proof
hide_const (open) ordering_constraint_proof'

definition
  id_labeling :: "(('f,'a list)lab,'a,'a list + ('f,'a list)lab list)label" 
where
  "id_labeling _ cs = Inl cs" 


definition scg_position :: "nat xmlt" where
  "scg_position \<equiv> Xml.nat ''position''"
hide_const (open) scg_position

definition
  scg :: "'f xmlt \<Rightarrow> (('f, string) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) xmlt"
where
  "scg xml2name \<equiv> xml1many_elements ''sizeChangeGraph'' 
    (CPF_Parser.rule xml2name)
    (xml3elements ''edge''
       CPF_Parser.scg_position
       (Xml.bool ''strict'')
       CPF_Parser.scg_position
       (\<lambda>p s q. (s,(p,q)))
    )
    (\<lambda>lr edges. (lr,(map snd (filter fst edges), map snd (filter (\<lambda> x. \<not> fst x) edges))))"
hide_const (open) scg

definition uncurry_info :: "'f xmlt \<Rightarrow> ('f, string) uncurry_info xmlt" where
  "uncurry_info xml2name \<equiv> xml4elements ''uncurryInformation''
    xml2name
    (xml_many_elements ''uncurriedSymbols''
      (xml2many_elements ''uncurriedSymbolEntry''
        xml2name
        (Xml.nat ''arity'')
        xml2name
        (\<lambda>f n fs. ((f, n), fs))) id)
    (xml1element ''uncurryRules'' (CPF_Parser.rules xml2name) id) 
    (xml1element ''etaRules'' (CPF_Parser.rules xml2name) id) 
    (\<lambda>a sml U E. (a, sml, U, E))"

definition innermostLhss :: "'a xmlt \<Rightarrow> ('a,string)term list xmlt" where
  "innermostLhss xml2name \<equiv> xml_many_elements ''innermostLhss'' 
            (CPF_Parser.term xml2name) id"

definition strategy :: "'a xmlt \<Rightarrow> ('a,string)strategy xmlt" where
  "strategy xml2name = xml1element ''strategy'' 
     (xml_options [
        (''innermost'', xml0elements ''innermost'' Innermost),
        (''innermostLhss'', xml_change (innermostLhss xml2name) Innermost_Q)
                  ]) id"

definition
  xml2dp_input :: "'f xmlt \<Rightarrow> (bool \<times> ('f, string) rules \<times> ('f,string)strategy \<times> ('f, string) rules) xmlt"
where
  "xml2dp_input xml2name \<equiv> 
   (xml_choice ''dpInput''
      [xml4elements ''dpInput'' 
         (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
         (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
         (strategy xml2name)  
         (Xml.bool ''minimal'')
         (\<lambda> r p s m. (m,p,s,r)),
       xml3elements ''dpInput'' 
         (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
         (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
         (Xml.bool ''minimal'')
         (\<lambda> r p m. (m,p,No_Strategy,r))
   ])" 

partial_function (parse_result)
  xml2dp_proof :: "('f, 'l) lab xmlt \<Rightarrow> ('f, 'l, string) dp_proof xmlt"
where
  [code]: "xml2dp_proof xml2name x = xml1element ''dpProof'' (xml_options [ 
    (''pIsEmpty'', xml0elements ''pIsEmpty'' PisEmpty),
    (''depGraphProc'', xml_many_elements ''depGraphProc''
      (xml2nd_choice ''component'' 
        (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
        ''realScc''
        (xml2dp_proof xml2name)
        (\<lambda>dps prfOpt. (prfOpt, dps)))
      DepGraphProc),
    (''redPairProc'', xml3elements ''redPairProc'' 
      (CPF_Parser.ordering_constraint_proof' xml2name) 
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
      (xml2dp_proof xml2name) 
      RedPairProc),
    (''usableRulesProc'', xml2elements ''usableRulesProc'' 
      (xml1element ''usableRules'' (CPF_Parser.rules xml2name) id) 
      (xml2dp_proof xml2name) 
      UsableRulesProc),
    (''innermostLhssRemovalProc'', xml2elements ''innermostLhssRemovalProc'' 
      (innermostLhss xml2name)
      (xml2dp_proof xml2name) 
      QReductionProc),
    (''rewritingProc'', xml4elements ''rewritingProc'' 
      (CPF_Parser.rule xml2name)
      (CPF_Parser.rstep xml2name)
      (xml1element ''usableRules'' (CPF_Parser.rules xml2name) id) 
      (xml2dp_proof xml2name) 
      (\<lambda> (s,t) (p,lr,t') U prof. RewritingProc U (s,t) (s,t') lr p prof)),
    (''narrowingProc'', xml4elements ''narrowingProc'' 
      (CPF_Parser.rule xml2name)
      CPF_Parser.pos
      (xml1element ''narrowings'' (CPF_Parser.rules xml2name) id) 
      (xml2dp_proof xml2name) 
      NarrowingProc),
    (''instantiationProc'', xml3elements ''instantiationProc'' 
      (CPF_Parser.rule xml2name)
      (xml1element ''instantiations'' (CPF_Parser.rules xml2name) id) 
      (xml2dp_proof xml2name) 
      InstantiationProc),
    (''forwardInstantiationProc'', xml3to4elements ''forwardInstantiationProc'' 
      (CPF_Parser.rule xml2name)
      (xml1element ''instantiations'' (CPF_Parser.rules xml2name) id) 
      (xml1element ''usableRules'' (CPF_Parser.rules xml2name) id) 
      (xml2dp_proof xml2name) 
      ForwardInstantiationProc),
    (''semlabProc'',xml_choice ''semlabProc'' [
     xml4elements ''semlabProc'' 
       (CPF_Parser.sl_variant xml2name) 
       (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
       (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
       (xml2dp_proof xml2name)
       (\<lambda>sli lp. SemlabProc sli lp []),
     xml5elements ''semlabProc'' 
       (CPF_Parser.sl_variant xml2name) 
       (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
       (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
       (innermostLhss xml2name)
       (xml2dp_proof xml2name)
       (\<lambda>sli lp lr lq. SemlabProc sli lp lq lr)
       ]),                    
    (''subtermProc'', xml1many2elements_gen ''subtermProc''
      (CPF_Parser.proj xml2name)
      (\<lambda>\<pi>. CPF_Parser.projected_rseq xml2name (create_proj \<pi>))
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
      (xml2dp_proof xml2name)
      SubtermCriterionProc),
    (''redPairUrProc'', xml4elements ''redPairUrProc'' 
      (CPF_Parser.ordering_constraint_proof' xml2name) 
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
      (xml1element ''usableRules'' (CPF_Parser.rules xml2name) id) 
      (xml2dp_proof xml2name)
      RedPairUrProc),
    (''monoRedPairUrProc'', xml5elements ''monoRedPairUrProc'' 
      (CPF_Parser.ordering_constraint_proof xml2name) 
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml1element ''usableRules'' (CPF_Parser.rules xml2name) id) 
      (xml2dp_proof xml2name)
      MonoRedPairUrProc),
    (''monoRedPairProc'', xml4elements ''monoRedPairProc'' 
      (CPF_Parser.ordering_constraint_proof xml2name) 
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml2dp_proof xml2name)
      MonoRedPairProc),
    (''uncurryProc'', xml_choice ''uncurryProc'' [
      xml4elements ''uncurryProc'' 
        (uncurry_info xml2name) 
        (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
        (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
        (xml2dp_proof xml2name)
        (UncurryProc None),
      xml5elements ''uncurryProc'' 
        (Xml.nat ''applicativeTop'')
        (uncurry_info xml2name) 
        (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
        (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
        (xml2dp_proof xml2name)
        (\<lambda>n. UncurryProc (Some n))
    ]),
    (''flatContextClosureProc'', xml5elements ''flatContextClosureProc''
      (xml1element ''freshSymbol'' xml2name id)
      (CPF_Parser.flat_contexts xml2name)
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2dp_proof xml2name)
      FccProc),
    (''switchInnermostProc'', xml2elements ''switchInnermostProc''
      (wcr_proof xml2name)
      (xml2dp_proof xml2name)
      SwitchInnermostProc),
    (''splitProc'', xml4elements ''splitProc'' 
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml2dp_proof xml2name)
      (xml2dp_proof xml2name)
      SplitProc),
    (''finitenessAssumption'', xml1element ''finitenessAssumption''
      (xml2dp_input xml2name)
      (\<lambda> (m,p,q,r). AssumeFinite (False,m,p, [], strategy_to_Q q r, [], r))),
    (''unlabProc'',xml3elements ''unlabProc''
         (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
         (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
         (xml2dp_proof xml2name)
         UnlabProc),
    (''sizeChangeProc'', xml1many_elements ''sizeChangeProc''
      (xml_options [
        (''subtermCriterion'', xml0elements ''subtermCriterion'' None),
        (''reductionPair'', xml1to2elements ''reductionPair'' 
          (CPF_Parser.ordering_constraint_proof xml2name) 
          (xml1element ''usableRules'' (CPF_Parser.rules xml2name) id) 
          (\<lambda>redp ur. Some (redp,ur)))
      ])
      (CPF_Parser.scg xml2name)
      (\<lambda>version. case version of
        None \<Rightarrow> SizeChangeSubterm
      | Some redp_ur \<Rightarrow> SizeChangeRedPair (fst redp_ur) (snd redp_ur))
  )]) id x"

definition xmlIgnore :: "string \<Rightarrow> unit xmlt" where
  "xmlIgnore name xml \<equiv> if name = (Xml.get_name xml)
    then return ()
    else error (''expected '' @ name @ '' element, but found '' @ to_string xml)"

definition bounds_type :: "boundstype xmlt" where
  "bounds_type \<equiv> xml1element ''type'' 
     (xml_options [
         (''roof'', xml0elements ''roof'' Roof),
         (''match'', xml0elements ''match'' Match)
     ])
     id"

definition bounds_bound :: "nat xmlt" where
  "bounds_bound \<equiv> Xml.nat ''bound''"

definition state :: "string xmlt" where
  "state = Xml.text ''state''"

definition final_states :: "string list xmlt" where
  "final_states = xml_many_elements ''finalStates'' state id"

definition transition_lhs :: "'a xmlt \<Rightarrow> (('a \<times> nat) \<times> string list + string) xmlt" where
  "transition_lhs xml2name x = (if (case x of (XML _ _ cs _) \<Rightarrow> length cs = 1) then 
    (xml1element ''lhs'' state Inr) x
   else
    xml2many_elements ''lhs''
    (xml2name)
    (Xml.nat ''height'')
    (state)
    (\<lambda>f h qs. Inl ((f, h), qs)) x)"

definition
  transition :: "'a xmlt \<Rightarrow> ((string, 'a \<times> nat) ta_rule + (string \<times> string)) xmlt"
where
  "transition xml2name = xml2elements ''transition''
    (transition_lhs xml2name)
    (xml1element ''rhs'' state id)
    (\<lambda>lhs q. case lhs of
      Inl (f, qs) \<Rightarrow> Inl (TA_rule f qs q)
    | Inr q' \<Rightarrow> Inr (q', q))"

definition
  transitions ::
    "'a xmlt \<Rightarrow> (((string, 'a \<times> nat) ta_rule list) \<times> ((string \<times> string) list)) xmlt"
where
  "transitions xml2name = xml_many_elements ''transitions'' (transition xml2name)
     (\<lambda>rls. let (rules, eps) = partition (\<lambda>rl. case rl of Inl _ \<Rightarrow> True | Inr _ \<Rightarrow> False) rls;
                        ruls = map (\<lambda>rl. case rl of Inl r \<Rightarrow> r) rules;
                        ep   = map (\<lambda>rl. case rl of Inr e \<Rightarrow> e) eps
            in (ruls, ep))"

definition bounds_tree_automaton :: "'a xmlt \<Rightarrow> (string, 'a \<times> nat) tree_automaton xmlt" where
  "bounds_tree_automaton xml2name \<equiv> xml2elements ''treeAutomaton''
    final_states
    (transitions xml2name)
    (\<lambda> f (t, eps). Tree_Automaton f t eps)"

definition bounds_info :: "'a xmlt \<Rightarrow> ('a, string) bounds_info xmlt" where
  "bounds_info xml2name \<equiv> xml4elements ''bounds''
     (bounds_type)
     (bounds_bound)
     final_states
     (bounds_tree_automaton xml2name)
     (Bounds_Info)"

type_synonym ('f,'v)i_o_strategy = "('f,'v)strategy + ('f,'v)ostrategy"

definition i_o_strategy :: "'a xmlt \<Rightarrow> ('a,string)i_o_strategy xmlt" where
  "i_o_strategy xml2name = xml1element ''strategy'' 
     (xml_options [
        (''innermost'', xml0elements ''innermost'' (Inl Innermost)),
        (''outermost'', xml0elements ''outermost'' (Inr Outermost)),
        (''innermostLhss'', xml_change (innermostLhss xml2name) (\<lambda> q. Inl (Innermost_Q q)))
                  ]) id"

definition
  xml2i_o_trs_input :: "'f xmlt \<Rightarrow> (('f,string)i_o_strategy \<times> ('f, string) rules \<times> ('f, string) rules option) xmlt"
where
  "xml2i_o_trs_input xml2name \<equiv> 
   (\<lambda> x. let e = no_transform ''trsInput'' x in (case x of XML tag [] (c # cs) None \<Rightarrow> if length cs \<le> 2 \<and> tag = ''trsInput'' then (
      do { 
         r \<leftarrow> (xml1element ''trs'' (CPF_Parser.rules xml2name) id) c;
         (str,cs') \<leftarrow> (if cs = [] \<or> Xml.get_name (hd cs) \<noteq> ''strategy'' then return (Inl No_Strategy, cs) else do { s \<leftarrow> (i_o_strategy xml2name) (hd cs); return (s,tl cs)});
         (s_opt,cs'') \<leftarrow> if (cs' = [] \<or> Xml.get_name (hd cs') \<noteq> ''relativeRules'') then return (None, cs') else do { s \<leftarrow> (xml1element ''relativeRules'' (CPF_Parser.rules xml2name) id) (hd cs'); return (Some s,tl cs')};
         if cs'' = [] then return (str,r,s_opt) else e
      }
   ) else e | _ \<Rightarrow> e))" 

definition
  xml2trs_input :: "'f xmlt \<Rightarrow> (('f,string)strategy \<times> ('f, string) rules \<times> ('f, string) rules option) xmlt"
where
  "xml2trs_input xml2name x \<equiv> do {
    (ios,r,s) <- xml2i_o_trs_input xml2name x;
    case ios of Inl is \<Rightarrow> return (is,r,s)
             |  Inr _ \<Rightarrow> error ''outermost strategy not supported at this point''
  }"

definition
  xml2otrs_input :: "'f xmlt \<Rightarrow> (('f,string)ostrategy \<times> ('f, string) rules) xmlt"
where
  "xml2otrs_input xml2name x \<equiv> do {
    (ios,r,s) <- xml2i_o_trs_input xml2name x;
    case (ios,s) of (Inr os,None) \<Rightarrow> return (os,r)
             | _ \<Rightarrow> error ''not an outermost strategy or relative rules present''
  }"

definition
  xml2completion_input :: "'f xmlt \<Rightarrow> (('f, string) equation list \<times> ('f, string) rules) xmlt"
where
  "xml2completion_input xml2name \<equiv> 
     xml2elements ''completionInput''
       (xml1element ''equations'' (CPF_Parser.rules xml2name) id)
       (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
       Pair"

definition
  xml2equational_input :: "'f xmlt \<Rightarrow> (('f, string) equation list \<times> ('f, string) term \<times> ('f, string) term) xmlt"
where
  "xml2equational_input xml2name \<equiv>
    xml2elements ''equationalReasoningInput''
      (xml1element ''equations'' (CPF_Parser.rules xml2name) id)
      (xml2elements ''equation''
        (CPF_Parser.term xml2name)
        (CPF_Parser.term xml2name)
        (Pair))
      Pair"

definition complexity_class :: "complexity_class xmlt"
  where "complexity_class \<equiv> 
       xml_change (Xml.nat ''polynomial'') Comp_Poly"

definition signature :: "'f xmlt \<Rightarrow> ('f \<times> nat) list xmlt"
  where "signature xml2name \<equiv> xml_many_elements ''signature'' 
    (xml2elements ''symbol'' xml2name (Xml.nat ''arity'') Pair) id"

definition complexity_measure :: "'f xmlt \<Rightarrow> ('f,string)complexity_measure xmlt"
  where "complexity_measure xml2name \<equiv> xml_options [
   (''derivationalComplexity'', xml1element ''derivationalComplexity'' (signature xml2name) Derivational_Complexity),
   (''runtimeComplexity'', xml2elements ''runtimeComplexity'' (signature xml2name) (signature xml2name) Runtime_Complexity)
  ]"

definition
  xml2complexity_input :: "'f xmlt \<Rightarrow> (('f,string)strategy \<times> ('f, string) rules \<times> ('f, string) rules option \<times> ('f,string)complexity_measure \<times> complexity_class) xmlt"
where
  "xml2complexity_input xml2name \<equiv> 
     xml3elements ''complexityInput''
       (xml2trs_input xml2name)
       (complexity_measure xml2name)
       complexity_class
       (\<lambda> (q,r,s) cm cc. (q,r,s,cm,cc))"

definition
  xml2input :: "'f xmlt \<Rightarrow>
    ((bool \<times> ('f, string) rules \<times> ('f, string) strategy \<times> ('f, string) rules)
    + (('f, string) strategy \<times> ('f, string) rules \<times> ('f, string) rules option)
    + (('f, string) equation list \<times> ('f, string) rules)
    + (('f, string) equation list \<times> ('f, string) term \<times> ('f, string) term)
    + (('f,string)strategy \<times> ('f, string) rules \<times> ('f, string) rules option \<times> ('f,string)complexity_measure \<times> complexity_class)
    + (('f, string) ostrategy \<times> ('f, string) rules)
    ) xmlt"
where
  "xml2input xml2name \<equiv> xml1element ''input'' (xml_choice ''input'' [
    xml_change (xml2dp_input xml2name) Inl,
    xml_change (xml2trs_input xml2name) (Inr \<circ> Inl),
    xml_change (xml2completion_input xml2name) (Inr \<circ> Inr \<circ> Inl),
    xml_change (xml2equational_input xml2name) (Inr \<circ> Inr \<circ> Inr \<circ> Inl),
    xml_change (xml2complexity_input xml2name) (Inr \<circ> Inr \<circ> Inr \<circ> Inr \<circ> Inl),
    xml_change (xml2otrs_input xml2name) (Inr \<circ> Inr \<circ> Inr \<circ> Inr \<circ> Inr)
  ]) id"

partial_function (parse_result)
  xml2trs_proof :: "('f, label_type) lab xmlt \<Rightarrow> ('f, label_type, string) trs_proof xmlt"
where
  [code]: "xml2trs_proof xml2name x = xml1element ''trsTerminationProof'' (xml_options [
    (''rIsEmpty'', xml0elements ''rIsEmpty'' RisEmpty),
    (''semlab'',xml_choice ''semlab'' [
     xml3elements ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
       (xml2trs_proof xml2name)
       (\<lambda>sli. Semlab sli []),
     xml4elements ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
       (innermostLhss xml2name)
       (xml2trs_proof xml2name)
       (\<lambda>sli lr lq. Semlab sli lq lr)
      ]),                    
    (''split'', xml3elements ''split'' 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml2trs_proof xml2name)
      (xml2trs_proof xml2name)
      Split),
    (''dpTrans'', xml3elements ''dpTrans'' 
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
      (Xml.bool ''markedSymbols'')
      (xml2dp_proof xml2name)
      (\<lambda>dps _ prf. DPTrans dps prf)),
    (''ruleRemoval'', xml3elements ''ruleRemoval''
      (CPF_Parser.ordering_constraint_proof xml2name) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2trs_proof xml2name)
      RuleRemoval),
    (''bounds'', (\<lambda>xml. bounds_info xml2name xml \<guillemotright>= (return \<circ> Bounds))),
    (''stringReversal'', xml2elements ''stringReversal''
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml2trs_proof xml2name)
      (\<lambda>_. StringReversal)),
    (''uncurry'', xml3elements ''uncurry'' 
      (uncurry_info xml2name)
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml2trs_proof xml2name)
      Uncurry),
    (''flatContextClosure'', xml3elements ''flatContextClosure''
      (CPF_Parser.flat_contexts xml2name)
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2trs_proof xml2name)
      Fcc),
    (''switchInnermost'', xml2elements ''switchInnermost''
      (wcr_proof xml2name)
      (xml2trs_proof xml2name)
      SwitchInnermost),
    (''terminationAssumption'', xml1element ''terminationAssumption''
        (xml2trs_input xml2name)
        (\<lambda> (q,r,so). AssumeSN (False,strategy_to_Q q r, r, case so of None \<Rightarrow> [] | Some s \<Rightarrow> s)))
  ]) id x"

partial_function (parse_result) 
  xml2complexity_proof :: "'f xmlt \<Rightarrow> ('f, string) complexity_proof xmlt"
where
  [code]: "xml2complexity_proof xml2name x = xml1element ''complexityProof'' (xml_options [
    (''ruleShifting'', xml3elements ''ruleShifting''
      (CPF_Parser.ordering_constraint_proof xml2name) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2complexity_proof xml2name)
      Rule_Shift_Complexity),
    (''rIsEmpty'', xml0elements ''rIsEmpty''
      RisEmpty_Complexity)
   ]) id x"


partial_function (parse_result)
  xml2reltrs_proof :: "('f, label_type) lab xmlt \<Rightarrow> ('f, label_type, string) trs_proof xmlt"
where
  [code]: "xml2reltrs_proof xml2name x = xml1element ''relativeTerminationProof'' (xml_options [
    (''rIsEmpty'', xml0elements ''rIsEmpty''
      RisEmpty),
    (''sIsEmpty'', xml1element ''sIsEmpty''
      (xml2trs_proof xml2name)
      id),
    (''ruleRemoval'', xml4elements ''ruleRemoval''
      (CPF_Parser.ordering_constraint_proof xml2name) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2reltrs_proof xml2name)
      (\<lambda>rp one two prf. RuleRemoval rp (one @ two) prf)),
    (''semlab'',xml_choice ''semlab'' [
      xml4elements ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
       (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
       (xml2reltrs_proof xml2name)
       (\<lambda>sli lr lrw. Semlab sli [] (lr @ lrw)),
      xml5elements ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
       (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
       (innermostLhss xml2name)
       (xml2reltrs_proof xml2name)
       (\<lambda>sli lr lrw lq. Semlab sli lq (lr @ lrw))
       ]),
    (''uncurry'', xml4elements ''uncurry'' 
      (uncurry_info xml2name)
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml2reltrs_proof xml2name)
      (\<lambda>i r s p. Uncurry i (r @ s) p)),
    (''stringReversal'', xml3elements ''stringReversal''
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml2reltrs_proof xml2name)
      (\<lambda> _ _. StringReversal)),
    (''equalityRemoval'',xml1element ''equalityRemoval'' 
       (xml2reltrs_proof xml2name)
       DropEquality),
    (''relativeTerminationAssumption'',
      xml1element ''relativeTerminationAssumption''
        (xml1to2elements ''trsInput''
          (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
          (xml1element ''relativeRules'' (CPF_Parser.rules xml2name) id)
          (\<lambda>r rwo. case rwo of
            None \<Rightarrow> AssumeSN (False,[], r, [])
          | Some rw \<Rightarrow> AssumeSN (False,[], r, rw))) id)
  ]) id x"

definition nonloop :: "'f xmlt \<Rightarrow> ('f,string) non_loop_prf xmlt"
  where "nonloop xml2name \<equiv> xml6elements ''nonLoop''
     (CPF_Parser.pat_rule_prf xml2name)
     (CPF_Parser.subst xml2name)
     (CPF_Parser.subst xml2name)
     (Xml.nat ''natural'')
     (Xml.nat ''natural'')
     CPF_Parser.pos
     Non_Loop_Prf
"



partial_function (parse_result)
  xml2dp_disproof :: "'f xmlt \<Rightarrow> ('f, string) dp_disproof xmlt"
where
  [code]: "xml2dp_disproof xml2name x = xml1element ''dpNonterminationProof'' (xml_options [
    (''dpRuleRemoval'', xml_choice ''dpRuleRemoval'' 
         [xml3elements ''dpRuleRemoval''
            (xml1element ''dps'' (CPF_Parser.rules xml2name) Some)
            (xml1element ''trs'' (CPF_Parser.rules xml2name) Some)
            (xml2dp_disproof xml2name)
               DP_Rule_Removal,
          xml2elements ''dpRuleRemoval''
            (xml1element ''trs'' (CPF_Parser.rules xml2name) Some)
            (xml2dp_disproof xml2name)
               (DP_Rule_Removal None),
          xml2elements ''dpRuleRemoval''
            (xml1element ''dps'' (CPF_Parser.rules xml2name) Some)
            (xml2dp_disproof xml2name)
               (\<lambda> p. DP_Rule_Removal p None)
         ]),
    (''loop'', xml_change (CPF_Parser.loop xml2name)
      (\<lambda>(s, rseq, \<sigma>, C). DP_Loop s rseq \<sigma> C)),
    (''nonLoop'', xml_change (CPF_Parser.nonloop xml2name)
      DP_Nonloop),
    (''innermostLhssRemovalProc'', xml2elements ''innermostLhssRemovalProc'' 
      (innermostLhss xml2name)
      (xml2dp_disproof xml2name) 
      DP_Q_Reduction),
    (''innermostLhssIncreaseProc'', xml2elements ''innermostLhssIncreaseProc'' 
      (innermostLhss xml2name)
      (xml2dp_disproof xml2name) 
      DP_Q_Increase),
    (''instantiationProc'', xml2elements ''instantiationProc'' 
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
      (xml2dp_disproof xml2name) 
      DP_Instantiation),
    (''switchFullStrategyProc'', xml2elements ''switchFullStrategyProc'' 
      (wcr_proof xml2name)
      (xml2dp_disproof xml2name) 
      DP_Termination_Switch)
  ]) id x"

partial_function (parse_result)
  xml2trs_disproof :: "'f xmlt \<Rightarrow> ('f, string) trs_disproof xmlt"
where
  [code]: "xml2trs_disproof xml2name x = xml1element ''trsNonterminationProof'' (xml_options [
  (''variableConditionViolated'', xml0elements ''variableConditionViolated''
    TRS_Not_Well_Formed),
  (''loop'', xml_change (CPF_Parser.loop xml2name)
    (\<lambda>(s, rseq, \<sigma>, C). TRS_Loop s (map (\<lambda>(x, y, _, z). (x, y, z)) rseq) \<sigma> C)),
  (''nonLoop'', xml_change (CPF_Parser.nonloop xml2name)
    TRS_Nonloop),
  (''ruleRemoval'', xml2elements ''ruleRemoval''
    (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
    (xml2trs_disproof xml2name)
    TRS_Rule_Removal),
  (''dpTrans'', xml3elements ''dpTrans''
    (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
    (Xml.bool ''markedSymbols'')
    (xml2dp_disproof xml2name)
    (\<lambda>p _ prf. TRS_DP_Trans p prf)),
  (''stringReversal'', xml2elements ''stringReversal''
    (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
    (xml2trs_disproof xml2name)
    (\<lambda>_. TRS_String_Reversal)),
  (''innermostLhssIncrease'', xml2elements ''innermostLhssIncrease'' 
      (innermostLhss xml2name)
      (xml2trs_disproof xml2name) 
      TRS_Q_Increase)
  ]) id x"

partial_function (parse_result)
  xml2otrs_disproof :: "'f xmlt \<Rightarrow> ('f, string) otrs_disproof xmlt"
where
  [code]: "xml2otrs_disproof xml2name x = xml1element ''trsNonterminationProof'' (xml_options [
  (''loop'', xml_change (CPF_Parser.loop xml2name)
    (\<lambda>(s, rseq, \<sigma>, C). OTRS_Loop s (map (\<lambda>(x, y, _, z). (x, y, z)) rseq) \<sigma> C)),
  (''ruleRemoval'', xml2elements ''ruleRemoval''
    (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
    (xml2otrs_disproof xml2name)
    OTRS_Rule_Removal)
  ]) id x"

partial_function (parse_result)
  xml2reltrs_disproof :: "'f xmlt \<Rightarrow> ('f, string) reltrs_disproof xmlt"
where
  [code]: "xml2reltrs_disproof xml2name x = xml1element ''relativeNonterminationProof'' (xml_options [
    (''variableConditionViolated'', xml0elements ''variableConditionViolated''
      Rel_Not_Well_Formed),
    (''loop'', xml_change (CPF_Parser.loop xml2name)
      (\<lambda>(s, rseq, \<sigma>, C). Rel_Loop s rseq \<sigma> C)),
    (''trsNonterminationProof'', xml_change (xml2trs_disproof xml2name) Rel_R_Not_SN),
    (''ruleRemoval'', xml3elements ''ruleRemoval''
      (xml1element ''trs'' (CPF_Parser.rules xml2name) Some)
      (xml1element ''trs'' (CPF_Parser.rules xml2name) Some)
      (xml2reltrs_disproof xml2name)
      Rel_Rule_Removal)
  ]) id x"

definition
  xml2cr_proof :: "('f,label_type)lab xmlt \<Rightarrow> ('f, label_type,string) cr_proof xmlt"
where
  "xml2cr_proof xml2name x = xml1element ''crProof'' (xml_options [
    (''wcrAndSN'', xml2elements ''wcrAndSN''
         (wcr_proof xml2name)
         (xml2trs_proof xml2name)
      SN_WCR),
    (''orthogonal'', xml0elements ''orthogonal'' Weakly_Orthogonal)
  ]) id x"

definition xml2nwcr_proof :: "unit xmlt"
  where "xml2nwcr_proof \<equiv> xml1element ''wcrDisproof'' (xml0elements ''nonJoinableCriticalPairsAuto'' ()) (\<lambda> _ . ())"

definition
  xml2ncr_proof :: "(string,label_type)lab xmlt \<Rightarrow> (string, label_type,string) ncr_proof xmlt"
where
  "xml2ncr_proof xml2name x \<equiv> let rew = CPF_Parser.rsteps xml2name in
    xml1element ''crDisproof'' (xml_options [
    (''nonWcrAndSN'', xml2elements ''nonWcrAndSN''
         xml2nwcr_proof
         (xml2trs_proof xml2name)
      (\<lambda> _. SN_NWCR)),
    (''nonJoinableFork'', xml3elements ''nonJoinableFork'' rew rew 
           (xml_options [
              (''distinctNormalForms'', xml0elements ''distinctNormalForms'' Diff_NFs),
              (''capNotUnif'', xml0elements ''capNotUnif'' Tcap_Not_Unif)
           ])
           (\<lambda> (s,seq1) (_,seq2). Non_Join s seq1 seq2))
  ]) id x"


definition
  xml2completion_proof :: "('f,label_type)lab xmlt \<Rightarrow> ('f, label_type,string) completion_proof xmlt"
where
  "xml2completion_proof xml2name = xml3elements ''completionProof'' 
     (wcr_proof xml2name)
     (xml2trs_proof xml2name)
     (xml1to2elements ''equivalenceProof''
       (CPF_Parser.subsumption_proof xml2name)
       (CPF_Parser.subsumption_proof xml2name)
       Pair)
     (\<lambda> w t (s1,s2). SN_WCR_Eq w t s1 s2)"

partial_function (parse_result) xml2eq_proof :: "'a xmlt \<Rightarrow> ('a, string) eq_proof xmlt" where
  [code]: "xml2eq_proof xml2name x = xml_options [
    (''refl'', xml1element ''refl'' (CPF_Parser.term xml2name) Refl),
    (''sym'', xml1element ''sym'' (xml2eq_proof xml2name) Sym),
    (''trans'', xml2elements ''trans'' (xml2eq_proof xml2name) (xml2eq_proof xml2name) Trans),
    (''assm'', xml2elements ''assm'' (CPF_Parser.rule xml2name) (CPF_Parser.subst xml2name) (\<lambda>r s. Assm r (mk_subst s))),
    (''cong'', xml1many_elements ''cong'' (xml2name) (xml2eq_proof xml2name) Cong)
  ] x"

definition
  xml2equational_proof :: "('f, label_type) lab xmlt \<Rightarrow> ('f, label_type, string) equational_proof xmlt"
where
  "xml2equational_proof xml2name =
    xml1element ''equationalProof'' (xml_options [
      (''equationalProofTree'', xml1element ''equationalProofTree'' (xml2eq_proof xml2name) Equational_Proof_Tree),
      (''conversion'', xml_change (CPF_Parser.conversion xml2name) Conversion),
      (''completionAndNormalization'', xml2elements ''completionAndNormalization''
        (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
        (xml2completion_proof xml2name)
        Completion_and_Normalization)
    ]) id"

definition
  xml2equational_disproof :: "('f, label_type) lab xmlt \<Rightarrow> ('f, label_type, string) equational_disproof xmlt"
where
  "xml2equational_disproof xml2name =
    xml1element ''equationalDisproof'' (xml_options [
      (''completionAndNormalization'', xml2elements ''completionAndNormalization''
        (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
        (xml2completion_proof xml2name)
        Completion_and_Normalization_Different)
    ]) id"

type_synonym lsymbol = "(string, label_type) lab"

fun
  xml2cert_problem :: "string xmlt \<Rightarrow> lsymbol xmlt \<Rightarrow> (string, label_type, string) cert_problem xmlt"
where
  "xml2cert_problem xml2uname xml2name (XML name egal cs text) =
  (if name = ''certificationProblem'' \<and> text = None \<and> length cs \<ge> 3 then do {
    inp     \<leftarrow> xml2input xml2name (cs ! 0);
    version \<leftarrow> Xml.text ''cpfVersion'' (cs!1);
    case inp of
      Inr (Inl (q,r,s_opt)) \<Rightarrow>
      (case s_opt of 
        None \<Rightarrow> xml1element ''proof'' (
        xml_options [
          (''trsTerminationProof'',
            xml_change (xml2trs_proof xml2name) (\<lambda>p. TRS_Proof q r None p)),
          (''crProof'',
            if strategy_to_Q q r = []
              then xml_change (xml2cr_proof xml2name) (\<lambda>p. TRS_Confluence r p)
              else (\<lambda> _. error ''strategies for confluence unsupported'')),
          (''crDisproof'',
            if strategy_to_Q q r = []
              then xml_change (xml2ncr_proof xml2name) (\<lambda>p. TRS_Non_Confluence r p)
              else (\<lambda> _. error ''strategies for confluence unsupported'')),
          (''trsNonterminationProof'',
            xml_change (xml2trs_disproof xml2name) (\<lambda>p. TRS_Disproof q r p))]
      ) id (cs!2)
      | Some s \<Rightarrow> xml1element ''proof'' (
        xml_options [
          (''relativeTerminationProof'',
            xml_change (xml2reltrs_proof xml2name) (\<lambda>p. TRS_Proof q r (Some s) p)),
          (''relativeNonterminationProof'', 
            xml_change (xml2reltrs_disproof xml2name) (\<lambda>p. Rel_TRS_Disproof q r s p))]
      ) id (cs!2))
    | Inl (m,p,q,r) \<Rightarrow> xml1element ''proof'' (
      xml_options [
        (''dpProof'',
          xml_change (xml2dp_proof xml2name) (\<lambda>prf. DP_Proof m p [] q [] r prf)),
        (''dpNonterminationProof'',
          xml_change (xml2dp_disproof xml2name) (\<lambda>prf. DP_Disproof m p q r prf))]
    ) id (cs!2)
    | Inr (Inr (Inl (e, r))) \<Rightarrow> xml1element ''proof'' (
      xml_options [
        (''completionProof'',
          xml_change (xml2completion_proof xml2name) (\<lambda>prf. Completion_Proof e r prf))]
    ) id (cs!2)
    | Inr (Inr (Inr (Inl (eqs, eq)))) \<Rightarrow> xml1element ''proof'' (
      xml_options [
        (''equationalProof'',
          xml_change (xml2equational_proof xml2name) (\<lambda>prf. Equational_Proof eqs eq prf)),
        (''equationalDisproof'',
          xml_change (xml2equational_disproof xml2name) (\<lambda>prf. Equational_Disproof eqs eq prf))
        ]
    ) id (cs!2)
    | Inr (Inr (Inr (Inr (Inl (q,r,s, cm, cc))))) \<Rightarrow> xml1element ''proof'' (
          xml_change (xml2complexity_proof xml2name) (\<lambda>prf. Complexity_Proof q r s cm cc prf)
    ) id (cs!2)
    | Inr (Inr (Inr (Inr (Inr (q,r))))) \<Rightarrow> xml1element ''proof'' (
          xml_change (xml2otrs_disproof xml2name) (\<lambda>prf. OTRS_Disproof q r prf)
    ) id (cs!2)
  } else error (''could not parse root element <'' @ name @ ''>''))"

partial_function (parse_result)
  plain_name :: "lsymbol xmlt"
where
  [code]: "plain_name x = xml_options [
     (''name'', xml_change (Xml.text ''name'') UnLab),
     (''sharp'', xml1element ''sharp'' plain_name Sharp),
     (''labeledSymbol'', xml2elements ''labeledSymbol'' plain_name (
           xml_options [
             (''numberLabel'', xml_many_elements ''numberLabel'' (Xml.nat ''number'') Inl),
             (''symbolLabel'', xml_many_elements ''symbolLabel'' plain_name Inr)
                       ])
     (\<lambda>f l. case l of
       Inl ls  \<Rightarrow> Lab f ls
     | Inr ls \<Rightarrow> FunLab f ls))
   ] x"

hide_const (open) plain_name

primrec
  xmldoc2cert_problem :: "xmldoc \<Rightarrow> (string,label_type,string)cert_problem sparse_result"
where
  "xmldoc2cert_problem (XMLDOC header xml) = debug
    ''0''
    ''parsing xml to certification problem''
    (xml2cert_problem (Xml.text ''name'') CPF_Parser.plain_name xml)"

fun
  parse_cert_problem :: "string \<Rightarrow> (string,label_type, string) cert_problem sparse_result"
where
  "parse_cert_problem s = (case doc_of_string s of Inl x \<Rightarrow> error x | Inr y \<Rightarrow> xmldoc2cert_problem y)"

declare parse_cert_problem.simps[simp del]

(* ... to xml *)
fun xml_lab :: "('f :: show, ('l :: show)list)lab \<Rightarrow> xml" where
  "xml_lab (UnLab x) = XML ''name'' [] [] (Some (to_string x))"
| "xml_lab (Sharp x) = XML ''sharp'' [] [xml_lab x] None"
| "xml_lab (FunLab x l) = XML ''labeledSymbol'' [] [xml_lab x, XML ''symbolLabel'' [] (map xml_lab l) None] None"
| "xml_lab (Lab x l) = XML ''labeledSymbol'' [] [xml_lab x, XML ''numberLabel'' [] (map (\<lambda> n. XML ''number'' [] [] (Some (to_string n))) l) None] None"

definition xml_signature :: "(('f :: show,('l :: show)list)lab \<times> nat)list \<Rightarrow> xml" where
  "xml_signature fs \<equiv> XML ''signature'' [] (map (\<lambda> (f,n). XML ''symbol'' [] [xml_lab f, XML ''arity'' [] [] (Some (shows n Nil))] None) fs) None"

fun xml_complexity_measure :: "(('f :: show,('l :: show)list)lab,'v)complexity_measure \<Rightarrow> xml" where
  "xml_complexity_measure (Derivational_Complexity F) = XML ''derivationalComplexity'' [] [xml_signature F] None"
 | "xml_complexity_measure (Runtime_Complexity C D) = XML ''runtimeComplexity'' [] [xml_signature C, xml_signature D] None"

fun xml_complexity_class :: "complexity_class \<Rightarrow> xml" where
  "xml_complexity_class (Comp_Poly n) = XML ''polynomial'' [] [] (Some (shows n Nil))"

fun xml_term :: "(('f :: show,('l :: show)list)lab, 'v :: show)term \<Rightarrow> xml" where
  "xml_term (Var x) = XML ''var'' [] [] (Some (to_string x))" 
| "xml_term (Fun f ts) = XML ''funapp'' [] 
    (xml_lab f # 
      (map (\<lambda> t. XML ''arg'' [] [xml_term t] None) ts)) 
      None"

definition xml_rules :: "string \<Rightarrow> (('f :: show,('l :: show)list)lab,'c :: show) rules \<Rightarrow> xml" where
  "xml_rules tag rls = XML tag Nil [XML ''rules'' [] (map (\<lambda> (l, r). XML ''rule'' [] [
    XML ''lhs'' [] [xml_term l] None,         
    XML ''rhs'' [] [xml_term r] None
  ] None) rls) None] None"

fun xml_Q :: "(('f :: show,('l :: show) list)lab,'v :: show)i_o_strategy \<Rightarrow> xml list"
  where "xml_Q (Inl No_Strategy) = []"
     |  "xml_Q (Inl Innermost) = [XML ''strategy'' [] [XML ''innermost'' [] [] None] None]"
     |  "xml_Q (Inr Outermost) = [XML ''strategy'' [] [XML ''outermost'' [] [] None] None]"
     |  "xml_Q (Inl (Innermost_Q Q)) = [XML ''strategy'' [] 
                                   [XML ''innermostLhss'' [] (map xml_term Q) None] None]"

fun xml_trs_input :: "(('f :: show,('l :: show) list)lab,'v :: show)i_o_strategy \<Rightarrow> (('f,'l list)lab,'v)rules
  \<Rightarrow> (('f,'l list)lab,'v)rules option \<Rightarrow> xml"
  where "xml_trs_input Q R None =  XML ''trsInput'' [] ([xml_rules ''trs'' R] @ xml_Q Q) None"
  |     "xml_trs_input Q R (Some S) = XML ''trsInput'' [] (xml_rules ''trs'' R # xml_Q Q @ [xml_rules ''relativeRules'' S]) None"

fun xml_cert_problem :: "('f :: show, ('l :: show) list, 'v :: show) cert_problem \<Rightarrow> xml" where
  "xml_cert_problem (TRS_Proof Q R S_o _) = xml_trs_input (Inl Q) R S_o"
| "xml_cert_problem (TRS_Disproof Q R _) = xml_trs_input (Inl Q) R None"
| "xml_cert_problem (OTRS_Disproof Q R _) = xml_trs_input (Inr Q) R None"
| "xml_cert_problem (Rel_TRS_Disproof Q R S _) = xml_trs_input (Inl Q) R (Some S)"
| "xml_cert_problem (DP_Proof m P Pw Q R Rw _) =
    XML ''dpInput'' [] (xml_rules ''trs'' Rw # xml_rules ''dps'' P # xml_Q (Inl Q) @ [XML ''minimal'' [] [] (Some (if m then ''true'' else ''false''))]) None"
| "xml_cert_problem (DP_Disproof m P Q R _) =
    XML ''dpInput'' [] (xml_rules ''trs'' R # xml_rules ''dps'' P # xml_Q (Inl Q) @ [XML ''minimal'' [] [] (Some (if m then ''true'' else ''false''))]) None"
| "xml_cert_problem (TRS_Confluence R _) = xml_trs_input (Inl No_Strategy) R None"
| "xml_cert_problem (TRS_Non_Confluence R _) = xml_trs_input (Inl No_Strategy) R None"
| "xml_cert_problem (Completion_Proof E R _) =
    XML ''completionInput'' [] [xml_rules ''equations'' E, xml_rules ''trs'' R] None"
| "xml_cert_problem (Equational_Proof E eq _) =
    XML ''equationalReasoningInput'' [] [xml_rules ''equations'' E, XML ''equation'' [] [xml_term (fst eq), xml_term (snd eq)] None] None"
| "xml_cert_problem (Equational_Disproof E eq _) =
    XML ''equationalReasoningInput'' [] [xml_rules ''equations'' E, XML ''equation'' [] [xml_term (fst eq), xml_term (snd eq)] None] None"
| "xml_cert_problem (Complexity_Proof Q R S cm cc _) =
    XML ''complexityInput'' [] [xml_trs_input (Inl Q) R S, xml_complexity_measure cm, xml_complexity_class cc] None"

end
