(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Innermost_Loops_Impl
imports
  Innermost_Loops 
  Trs_Impl
  Transitive_Closure_List_Impl
  Q_Restricted_Rewriting
begin

context fixed_subst
begin

lemma W_via_rtrancl: 
  "W t = {(x, y). x \<noteq> y \<and> y \<in> vars_term (\<mu> x)}^* `` vars_term t"
proof -
  obtain V where V: "V = vars_term t" by auto
  let ?M = "\<lambda> i V. \<Union>vars_term ` (\<mu> ^^ i) ` V"
  let ?R = "{(x,y) | x y. y \<in> vars_term (\<mu> x)}"
  have "?R^* `` V = (\<Union>n. (?R ^^ n) `` V)"
    unfolding rtrancl_is_UN_relpow by auto
  also have "... = (\<Union> n. ?M n V)"
  proof -
    {
      fix n 
      have "?M n V = (?R ^^ n) `` V"
      proof (induct n)
        case 0
        show ?case by auto
      next
        case (Suc n)
        have "(?R ^^ Suc n) `` V = (?R ^^ n O ?R) `` V" by auto
        also have "... = ?R `` ((?R ^^ n) `` V)" by auto
        also have "... = ?R `` ?M n V" unfolding Suc ..
        also have "... = ?M (Suc n) V" 
        proof -
          have "?M (Suc n) V = vars_term (t \<cdot> \<mu> ^^ Suc n)" 
            unfolding V vars_term_subst ..
          also have "... = vars_term (t \<cdot> \<mu> ^^ n \<cdot> \<mu>)" 
            unfolding subst_power_Suc by auto
          also have "... = ?R `` ?M n V" 
            unfolding V vars_term_subst by auto
          finally show ?thesis by simp
        qed
        finally show ?case by simp
      qed
    } note main = this
    show ?thesis unfolding main ..
  qed
  also have "... = W t" unfolding W_def vars_term_subst V by auto
  finally have main: "W t = ?R^* `` V" by simp
  show ?thesis unfolding main rtrancl_r_diff_Id[of ?R, symmetric] unfolding V[symmetric] 
    by (rule arg_cong[where f = "\<lambda> x. x^* `` V"], auto)
qed

lemma V_incr_via_rtrancl:
  "V_incr = {(y, x) | x y. x \<noteq> y \<and> \<mu> x = Var y}^* `` {x. is_Fun (\<mu> x)}"
  (is "_ = ?r^* `` ?s")
proof -
  {
    fix x y n
    assume "(x,y) \<in> ?r^*" 
    then obtain n where "(x,y) \<in> ?r^^n" unfolding rtrancl_is_UN_relpow ..
    hence "(\<mu> ^^ n) y = Var x"
    proof (induct n arbitrary: y)
      case 0 thus ?case by force
    next
      case (Suc n y)
      from Suc(2) obtain z where xz: "(x,z) \<in> ?r ^^ n" and zy: "(z,y) \<in> ?r" by auto
      from zy have zy: "\<mu> y = Var z" by auto
      from Suc(1)[OF xz] have x: "Var x = (\<mu> ^^ n) z" by simp
      show ?case unfolding x substpow.simps
        subst_compose_def zy by simp
    qed
    hence "\<exists> n. (\<mu> ^^ n) y = Var x" by auto
  } note one_dir = this
  {
    fix x
    assume "x \<in> ?r^* `` ?s"
    then obtain y where yx: "(y,x) \<in> ?r^*" and y: "y \<in> ?s" by auto
    from one_dir[OF yx] obtain n where x: "(\<mu> ^^ n) x = Var y" ..
    have "x \<in> V_incr" unfolding V_incr_def
      by (rule, rule exI[of _ "Suc n"], 
        unfold subst_power_Suc, insert y, simp add: x)
  }
  moreover 
  {
    fix x
    let ?N = "\<lambda> x n. is_Fun ((\<mu> ^^ n) x)"
    let ?P = "?N x"
    let ?Q = "\<lambda> x. \<exists> n. ?N x n"
    assume "x \<in> V_incr"
    then obtain n where nvar: "?P n" unfolding V_incr_def by auto
    let ?n = "LEAST n. ?P n"
    have nle: "?P ?n" by (rule LeastI, rule nvar)
    from Least_Suc[of ?P, OF nvar] have "?n > 0" by auto
    then obtain n where n: "?n = Suc n" by (cases ?n, auto)
    hence "n < ?n" by auto
    from not_less_Least[OF this] obtain y where y: "(\<mu> ^^ n) x = Var y"
      by (cases "(\<mu> ^^ n) x", auto)
    have "(\<mu> ^^ ?n) x = \<mu> y" unfolding n subst_power_Suc subst_compose_def y
      by simp
    with nle have ys: "y \<in> ?s" by simp
    have "?Q y" by (rule exI[of _ "Suc 0"], insert ys, auto)
    from y have "(y,x) \<in> ?r^^n" 
    proof (induct n arbitrary: x)
      case (0 x)
      thus ?case by auto
    next
      case (Suc n x)
      note id = Suc(2)[unfolded substpow.simps subst_compose_def]
      from id obtain z where xz: "\<mu> x = Var z" and zy: "(\<mu> ^^ n) z = Var y"
        by (cases "\<mu> x", auto)
      from Suc(1)[OF zy] have "(y,z) \<in> ?r ^^ n" .
      moreover have "(z,x) \<in> ?r"
      proof(rule, intro exI, rule conjI[OF refl], rule conjI[OF _ xz])
        show "x \<noteq> z"
        proof
          assume "x = z"
          with xz id have x: "\<mu> x = Var x" and xy: "(\<mu> ^^ n) x = Var y" by auto
          {
            fix n
            have "(\<mu> ^^ n) x = Var x" 
              by (induct n, insert x, auto simp: subst_compose_def)
          } note x = this
          have "\<mu> y = Var y \<cdot> \<mu>" by simp
          also have "... = (\<mu> ^^ (Suc n)) x" unfolding xy[symmetric]
            subst_power_Suc subst_compose_def ..
          also have "... = Var x" unfolding x ..
          finally show False using ys by auto
        qed
      qed
      ultimately show ?case by auto
    qed
    hence "(y,x) \<in> ?r^*" unfolding rtrancl_is_UN_relpow by auto
    with ys have "x \<in> ?r^* `` ?s" by auto
  }
  ultimately show ?thesis by blast
qed
end


locale fixed_subst_impl = 
  fixes \<mu> :: "('f, 'v) substL"

sublocale fixed_subst_impl \<subseteq> fixed_finite_subst "mk_subst \<mu>"  
  by (unfold_locales, rule finite_mk_subst)

definition V_incr_impl where
  "V_incr_impl d \<equiv>
    let 
      (ixs, uxs) = partition (\<lambda>(x, t). is_Fun t) d;
      rel = map (\<lambda>(x, y). (the_Var y, x)) uxs;
      start = map fst ixs
    in rtrancl_list_impl rel start"

definition W_impl :: "('f, 'v) substL \<Rightarrow> ('f, 'v) term \<Rightarrow> 'v list" where
  "W_impl d \<equiv>
    let 
      filt = filter (\<lambda>(x,y). x \<noteq> y);
      xvs = concat (map (\<lambda> (x,t). (map (\<lambda> y. (x,y)) (vars_term_list t))) d);
      rel = filt xvs;
      rtran = rtrancl_list_impl rel
    in (\<lambda>t. rtran (vars_term_list t))"

context fixed_subst_impl
begin

     
lemma W_impl: "set (W_impl (mk_subst_domain \<mu>) t) = W t" 
proof -  
  let ?I = "set [(x,y)\<leftarrow>concat (map (\<lambda>(x, t). map (Pair x) (vars_term_list t))
        (mk_subst_domain \<mu>)) . x \<noteq> y]"
  let ?S = " {(x, y) |x y. x \<noteq> y \<and> y \<in> vars_term (mk_subst \<mu> x)}"
  have id: "?I = ?S" 
  proof -
    have "?I \<subseteq> ?S" unfolding set_concat set_map set_filter mk_subst_domain
      by auto
    moreover {
      fix xy
      assume S: "xy \<in> ?S"
      obtain x y where xy: "xy = (x,y)" by (cases xy, auto)
      with S have neq: "x \<noteq> y" and y: "y \<in> vars_term (mk_subst \<mu> x)" by auto
      hence "mk_subst \<mu> x \<noteq> Var x" by (cases "mk_subst \<mu> x", auto)
      hence x: "x \<in> subst_domain (mk_subst \<mu>)" unfolding subst_domain_def by auto
      have "(x,y) \<in> ?I"
        unfolding set_filter set_concat set_map split mk_subst_domain
        by (rule, rule, rule, rule, rule, rule refl, rule x, simp_all add: y neq)
      hence "xy \<in> ?I" unfolding xy .
    }
    ultimately show "?I = ?S" by blast
  qed
  show ?thesis
    unfolding W_via_rtrancl
    unfolding W_impl_def Let_def
    unfolding rtrancl_list_impl
    unfolding id vars_term_list_sound by auto
qed

lemma V_incr_impl: "set (V_incr_impl (mk_subst_domain \<mu>)) = V_incr"
proof -
  let ?V = "V_incr_impl (mk_subst_domain \<mu>)"
  let ?\<mu> = "mk_subst \<mu>"
  let ?R = "(\<lambda>(x, y). (the_Var y, x)) `
              {x \<in> (\<lambda>x. (x, mk_subst \<mu> x)) ` subst_domain (mk_subst \<mu>).
               \<not> (case x of (x, t) \<Rightarrow> is_Fun t)}"
  let ?r = "{(y,x) | x y. x \<noteq> y \<and> ?\<mu> x = Var y}"
  let ?S = "fst ` {x \<in> (\<lambda>x. (x, mk_subst \<mu> x)) ` subst_domain (mk_subst \<mu>).
            case x of (x, t) \<Rightarrow> is_Fun t}"
  let ?s = "{ x. is_Fun (?\<mu> x)}"
  have idR: "?R = ?r" unfolding subst_domain_def by force
  have idS: "?S = ?s" unfolding subst_domain_def by force
  have "set ?V = ?r^* `` ?s"
    unfolding V_incr_impl_def Let_def
    unfolding partition_filter_conv split o_def 
    unfolding rtrancl_list_impl set_map set_filter mk_subst_domain
    unfolding split
    unfolding idS idR by auto
  also have "... = V_incr" unfolding V_incr_via_rtrancl ..
  finally show ?thesis .
qed
end

declare fixed_subst.conflicts.simps[code]

partial_function (option)
  ident_solve'_impl ::
    "('f, 'v) subst \<Rightarrow> ('f, 'v)ident_prob list \<Rightarrow> ('f, 'v) nident_prob \<Rightarrow> nat option"
where
  [code]: "ident_solve'_impl \<mu> cps stn = (let cp = remdups (fixed_subst.conflicts \<mu> stn) in 
    if Bex (set cp) (\<lambda>(u, v, m). is_Fun u) then None
    else if Bex (set cp) (\<lambda>(u, v, m). (u, v) \<in> set cps) then None
    else do {
      is \<leftarrow> mapM (\<lambda>(u, v, m). ident_solve'_impl \<mu> ((u, v) # cps) (u \<cdot> \<mu>, v, Suc m)) cp;
      Some (max_list (map Suc is))
    })"

definition ident_solve_impl :: "('f, 'v) subst \<Rightarrow> ('f, 'v) ident_prob \<Rightarrow> nat option"  where
  "ident_solve_impl \<equiv> \<lambda>\<mu> (s, t). ident_solve'_impl \<mu> [] (s, t, 0)"

context fixed_subst_impl
begin

lemma ident_solve'_impl:
  "ident_solve'_impl (mk_subst \<mu>) [] = ident_solve' {}"
proof -
  {
    fix cps stn
    let ?\<mu> = "mk_subst \<mu>"
    let ?map = "\<lambda> (u :: ('f,'v)term ,v :: ('f,'v)term , m :: nat). (u,v)"
    let ?eq = "\<lambda> stn cps cps'. ident_solve'_impl ?\<mu> cps' stn = ident_solve' cps stn"
    let ?p = "\<lambda> cps stn. \<forall> cps'. ?map ` cps = set cps' \<longrightarrow> ?eq stn cps cps'"
    have "?p cps stn" 
    proof (induct rule: ident_solve'.induct[of ?p])
      case (1 cps stn)
      show ?case
      proof (intro allI impI)
        fix cps'
        assume cps: "?map ` cps = set cps'"
        let ?cp = "set (conflicts stn)"
        note simpl = ident_solve'.simps[of cps stn, unfolded cps Let_def]
        note simpr = ident_solve'_impl.simps[of ?\<mu> cps' stn, unfolded Let_def V_incr_impl]
        note simpr = simpr[unfolded set_remdups set_map substpow_impl]
        let ?l1 = "\<exists> f ss t n. (Fun f ss, t,n) \<in> ?cp"
        let ?r1 = "\<exists> (u,v,n) \<in> ?cp. is_Fun u"
        have id1: "?l1 = ?r1" by force
        let ?l4 = "\<exists> u v n m. (u,v,n) \<in> ?cp \<and> (u,v,m) \<in> cps"
        let ?r4 = "\<exists> (u,v,m) \<in> ?cp. (u,v) \<in> set cps'"
        have id4: "?l4 = ?r4" using cps by force
        note simp = simpl[unfolded id1 id4] simpr
        have "?eq stn cps cps'"
        proof (cases ?r1)
          case True
          thus ?thesis unfolding simp by auto
        next
          case False hence l1: "\<not> ?l1" using id1 by simp
          from False have "?r1 = False" by simp
          note simp = simp[unfolded this]
          show ?thesis
          proof (cases ?r4)
            case True
            thus ?thesis unfolding simp by simp
          next
            case False hence l4: "\<not> ?l4" using id4 by simp
            from False have "?r4 = False" by simp
            note simp = simp[unfolded this]
            let ?max = "\<lambda> is. Some (max_list (map (\<lambda> i. Suc i) is))"
            let ?g = "(\<lambda>(u,v,m). ident_solve' (insert (u,v,m) cps) (u \<cdot> ?\<mu>, v, Suc m))"
            let ?st = "conflicts stn"
            let ?rst = "remdups ?st"
            let ?recl = "mapM ?g ?st"
            have l: "ident_solve' cps stn = ?recl \<guillemotright>= ?max" unfolding simp by simp
            let ?f = "(\<lambda> (u,v,m). ident_solve'_impl ?\<mu> ((u,v) # cps') (u \<cdot> ?\<mu>, v, Suc m))"
            let ?recr = "mapM ?f ?rst"
            have r: "ident_solve'_impl ?\<mu> cps' stn = ?recr \<guillemotright>= ?max" unfolding simp by simp
            note cpss = cps[symmetric]
            note ind = 1[OF refl l1 l4 _ refl refl, rule_format, unfolded image_insert split cps set.simps(2)[symmetric], OF _ refl] 
            hence id1: "set (map ?f ?rst) = set (map ?g ?st)" by force
            hence id2: "(\<forall> x \<in> set (map ?f ?rst). x \<noteq> None) = (\<forall> x \<in> set (map ?g ?st). x \<noteq> None)" by auto
            show ?thesis
            proof (cases ?recl)
              case None
              hence "?recr = None" unfolding mapM_None using id1 by force
              thus ?thesis using None unfolding l r by simp
            next
              case (Some iss)
              note mapM1 = mapM_Some[THEN conjunct1]
              obtain iss' where recr: "?recr = Some iss'" 
              proof (cases ?recr)
                case (Some iss') thus ?thesis ..
              next
                case None
                from mapM_Some[OF Some, THEN conjunct2] None[unfolded mapM_None] id2 show ?thesis by auto
              qed
              show ?thesis unfolding l r recr Some mapM1[OF Some] mapM1[OF recr] bind_lunit option.simps
                by (rule max_list_eq_set, unfold set_map, rule arg_cong[where f = "\<lambda> x. Suc ` x"], 
                  insert arg_cong[OF id1, of "\<lambda> x. the ` x"],
                  unfold set_map set_remdups, blast)
            qed                
          qed
        qed
        thus "?eq stn cps cps'" unfolding cps .
      qed
    qed
  } note main = this
  show ?thesis 
    by (rule ext, insert main[rule_format, of "set []" "[]"], simp)
qed

lemma ident_solve_impl:
  "ident_solve_impl (mk_subst \<mu>) = ident_solve"
  unfolding ident_solve_impl_def ident_solve'_impl ident_solve_def ..
end

definition ident_decision where "ident_decision ident ip \<equiv> case ident_solve_impl ident ip of None \<Rightarrow> False | Some _ \<Rightarrow> True"

lemma (in fixed_subst_impl) ident_decision[simp]: "ident_decision (mk_subst \<mu>) st = (\<exists> sol. ident_solution st sol)"
  unfolding ident_decision_def ident_solve_impl using ident_solve[of st] by force

partial_function (option)
  simplify_mp_impl ::
    "('f, 'v) subst \<Rightarrow> 'v list \<Rightarrow> ('f, 'v) gmatch_prob \<Rightarrow> ('f, 'v) gmatch_prob \<Rightarrow> (('f, 'v) gmatch_prob \<times> nat) option"
where
  [code]: "simplify_mp_impl \<mu> V_incr mp1 solved =
  (case mp1 of
    [] \<Rightarrow> Some (solved, 0)
  | ((s, Var x) # mp) \<Rightarrow> simplify_mp_impl \<mu> V_incr mp ((s, Var x) # solved)
  | ((Fun g ts, Fun f ls) # mp) \<Rightarrow> do {
    guard (f = g);
    pairs \<leftarrow> zip_option ts ls;
    simplify_mp_impl \<mu> V_incr (pairs @ mp) solved
  }
  | ((Var x, Fun f ls) # mp) \<Rightarrow> do {
    guard (x \<in> set V_incr);
    let m = map (\<lambda>(s, l). (s \<cdot> \<mu>, l));
    (smp, i) \<leftarrow> simplify_mp_impl \<mu> V_incr (m mp1) (m solved);
    Some (smp, Suc i)
   })"


context fixed_subst_impl
begin

lemma simplify_mp_impl: "simplify_mp = simplify_mp_impl (mk_subst \<mu>) (V_incr_impl (mk_subst_domain \<mu>))" (is "?S = ?I")
proof (intro ext)
  let ?\<mu> = "mk_subst \<mu>"
  let ?V = "V_incr_impl (mk_subst_domain \<mu>)"
  note sI = simplify_mp_impl.simps[of ?\<mu> ?V]
  note sS = simplify_mp.simps
  fix mp1 mp2
  show "?S mp1 mp2 = ?I mp1 mp2"
  proof (induct rule: simplify_mp.induct)
    case (1 mp2)
    show ?case unfolding sI[of Nil] by simp
  next
    case (2 s x mp1 mp2)
    let ?mp1 = "((s,Var x) # mp1)"
    show ?case unfolding sI[of ?mp1] sS unfolding 2 by (cases s, auto)
  next
    case (3 g ts f ls mp1 mp2)
    let ?mp1 = "((Fun g ts, Fun f ls) # mp1)"
    show ?case unfolding sI[of ?mp1] sS
      by (cases "f = g", cases "zip_option ts ls", insert 3) auto
  next
    case (4 x f ls mp1 mp2)
    let ?mp1 = "((Var x, Fun f ls) # mp1)"
    show ?case unfolding sI[of ?mp1] sS Let_def V_incr_impl
      by (cases "x \<in> V_incr", insert 4, auto)
  qed
qed
end
      
definition
  simplify_mp ::
    "('f, 'v) substL \<Rightarrow> 'v list \<Rightarrow> ('f, 'v) gmatch_prob \<Rightarrow> ('f, 'v) gmatch_prob \<Rightarrow> (('f, 'v) gmatch_prob \<times> nat) option"
where
  "simplify_mp \<mu> V \<equiv> let \<mu>' = mk_subst \<mu> in simplify_mp_impl \<mu>' V"

lemma simplify_mp: "simplify_mp \<mu> (V_incr_impl (mk_subst_domain \<mu>)) = fixed_subst.simplify_mp (mk_subst \<mu>)"
  unfolding simplify_mp_def Let_def fixed_subst_impl.simplify_mp_impl ..

definition gmatch_to_idents_impl where
  "gmatch_to_idents_impl simp ident mp \<equiv> do {
    (smp,i) \<leftarrow> simp mp [];
    let ids = map ident (ident_prob_of_smp smp);
    if None \<in> set ids then None else Some (max_list (map the ids) + i)
  }"

context fixed_finite_subst
begin
lemma gmatch_to_idents_impl_complete:
  assumes "gmatch_solution mp (n,\<sigma>)"
  shows "gmatch_to_idents_impl simplify_mp ident_solve mp \<noteq> None"
proof -
  from gmatch_to_idents_complete[OF assms] 
  obtain i idps where res1: "gmatch_to_idents mp = Some (i,idps)" and 
    sol: "\<And> idp. idp \<in> set idps \<Longrightarrow> \<exists> n. ident_solution idp n" by auto
  from ident_solve(2) sol have sol: "\<And> idp. idp \<in> set idps \<Longrightarrow> ident_solve idp \<noteq> None" by force
  from res1[unfolded gmatch_to_idents_def]
  obtain  smp where res: "local.simplify_mp mp [] = Some (smp, i)" "idps = ident_prob_of_smp smp"
    by (cases "local.simplify_mp mp []", auto)
  show ?thesis unfolding gmatch_to_idents_impl_def using sol ident_solve(2) unfolding res by force
qed

lemma gmatch_to_idents_impl_sound:
  assumes "gmatch_to_idents_impl simplify_mp ident_solve mp = Some i"
  shows "\<exists> \<sigma>. gmatch_solution mp (i,\<sigma>)"
proof -
  note sol = assms[unfolded gmatch_to_idents_impl_def]
  from sol
  obtain j smp where res1: "local.simplify_mp mp [] = Some (smp,j)"
    by (cases "local.simplify_mp mp []", auto)
  note sol = sol[unfolded res1]
  let ?ids = "map ident_solve (ident_prob_of_smp smp)"
  from sol have none: "None \<notin> set ?ids" by auto
  with sol have i: "i = max_list (map the ?ids) + j" by simp
  have match: "gmatch_to_idents mp = Some (j,ident_prob_of_smp smp)"
    unfolding gmatch_to_idents_def unfolding res1 by simp
  {
    fix idp
    assume "idp \<in> set (ident_prob_of_smp smp)"
    with none have "ident_solve idp \<noteq> None" by force
    then obtain j where id: "ident_solve idp = Some j" by auto
    from ident_solve(1)[OF id]
    have "ident_solution idp (the (ident_solve idp))" unfolding id by simp
  }      
  from gmatch_to_idents_sound[OF match this] obtain \<sigma> where "gmatch_solution mp (i,\<sigma>)" unfolding i map_map o_def by auto
  thus ?thesis by blast
qed
end

definition match_to_idents_impl where
  "match_to_idents_impl simp ident t_l \<equiv> gmatch_to_idents_impl simp ident [t_l]"

context fixed_finite_subst
begin
lemma match_to_idents_impl_complete:
  assumes "match_solution t_l (n,\<sigma>)"
  shows "match_to_idents_impl simplify_mp ident_solve t_l \<noteq> None"
  using gmatch_to_idents_impl_complete[of "[t_l]" n \<sigma>] assms
  unfolding match_to_idents_impl_def gmatch_solution_def by auto

lemma match_to_idents_impl_sound:
  assumes "match_to_idents_impl simplify_mp ident_solve t_l = Some i"
  shows "\<exists> \<sigma>. match_solution t_l (i,\<sigma>)"
  using gmatch_to_idents_impl_sound[of "[t_l]" i] assms
  unfolding match_to_idents_impl_def gmatch_solution_def by auto
end

fun match_prob_of_rp_impl where
  "match_prob_of_rp_impl \<mu> W (t, Var x) = [(t, Var x)]"
| "match_prob_of_rp_impl \<mu> W (t, l) = (
    let
      sterms = remdups (t # map \<mu> (W t));
      uterms = concat (map (filter is_Fun \<circ> supteq_list) sterms)
    in map (\<lambda>u. (u, l)) (remdups uterms))"

context fixed_subst_impl
begin

lemma match_prob_of_rp_impl:
  "set (match_prob_of_rp_impl (mk_subst \<mu>) (W_impl (mk_subst_domain \<mu>)) (t,l)) = match_prob_of_rp (t,l)"
proof (cases l)
  case (Var x)
  show ?thesis unfolding match_prob_of_rp_def Var by simp
next
  case (Fun f ls)
  show ?thesis unfolding Fun match_prob_of_rp_def match_prob_of_rp_impl.simps
    unfolding Let_def using W_impl
    by auto
qed
end

definition gmatch_decision where 
  "gmatch_decision \<mu> simp ident mp \<equiv>
    gmatch_to_idents_impl simp ident mp \<noteq> None"

lemma (in fixed_subst_impl) gmatch_decision:
  "gmatch_decision (mk_subst \<mu>) simplify_mp ident_solve mp = (\<exists>sol. gmatch_solution mp sol)"
  (is "?l = ?r")
proof -
  let ?\<mu> = "mk_subst \<mu>"
  show ?thesis
  proof
    assume ?r
    then obtain i \<sigma> where sol: "gmatch_solution mp (i,\<sigma>)" by auto
    show ?l unfolding gmatch_decision_def Let_def 
      by (rule gmatch_to_idents_impl_complete[OF sol])
  next
    assume ?l
    from this[unfolded gmatch_decision_def Let_def]
    obtain i where sol: "gmatch_to_idents_impl simplify_mp ident_solve mp = Some i" by auto
    from gmatch_to_idents_impl_sound[OF sol] show ?r by blast
  qed
qed

definition match_decision where 
  "match_decision \<mu> simp ident mp \<equiv> gmatch_decision \<mu> simp ident [mp]"

lemma (in fixed_subst_impl) match_decision:
  "match_decision (mk_subst \<mu>) simplify_mp ident_solve mp = (\<exists>sol. match_solution mp sol)"
  unfolding match_decision_def gmatch_decision using gmatch_solution_def by auto

definition redex_decision where
  "redex_decision \<mu> simp ident W rp \<equiv>
    let
      mps = match_prob_of_rp_impl \<mu> W rp;
      solver = match_decision \<mu> simp ident
    in (\<exists> mp \<in> set mps. solver mp)"

context fixed_subst_impl
begin

lemma redex_decision:
  "redex_decision (mk_subst \<mu>) simplify_mp ident_solve (W_impl (mk_subst_domain \<mu>)) rp = (\<exists>sol. redex_solution rp sol)"
  (is "?l = ?r")
proof -
  let ?\<mu> = "mk_subst \<mu>"
  obtain t l where rp: "rp = (t,l)" by force
  show ?thesis
  proof
    assume ?r
    then obtain i \<sigma> C where "redex_solution (t,l) (i,\<sigma>,C)" unfolding rp by auto
    from match_prob_of_rp_complete[OF this]
    obtain \<tau> u j
      where mem: "(u,l) \<in> match_prob_of_rp (t,l)"
      and sol: "match_solution (u,l) (j,\<tau>)" by auto
    show ?l unfolding rp redex_decision_def Let_def 
      set_map match_prob_of_rp_impl
      using match_decision[of "(u,l)"] sol mem by auto
  next
    assume ?l
    from this[unfolded rp redex_decision_def Let_def set_map match_prob_of_rp_impl]
    obtain m p where mp: "(m,p) \<in> match_prob_of_rp (t,l)" and sol: "match_decision (mk_subst \<mu>) simplify_mp ident_solve (m,p)" by auto
    from sol[unfolded match_decision] match_prob_of_rp_sound[OF _ mp] show ?r unfolding rp by force
  qed
qed
end

definition redex_rps_decision :: "('f, 'v) substL \<Rightarrow> ('f, 'v) redex_prob list \<Rightarrow> ('f, 'v) redex_prob check" where
  "redex_rps_decision \<mu> \<equiv> 
    let
      \<mu>' = mk_subst \<mu>;
      d = mk_subst_domain \<mu>;
      V = V_incr_impl d;
      simp = simplify_mp \<mu> V;
      ident = ident_solve_impl \<mu>';
      W = W_impl d;
      main = redex_decision \<mu>' simp ident W
    in check_allm (\<lambda>tl. check (\<not> main tl) tl)"

lemma redex_rps_decision:
  "isOK (redex_rps_decision \<mu> rps) = (\<not>(\<exists>t l i C \<sigma>. (t, l) \<in> set rps \<and> t \<cdot> ((mk_subst \<mu>) ^^ i) = C\<langle>l \<cdot> \<sigma>\<rangle>))"
  (is "?l = ?r")
proof -
  interpret fixed_subst_impl \<mu> .  
  show ?thesis unfolding redex_rps_decision_def Let_def 
    unfolding ident_solve_impl simplify_mp redex_decision 
    using redex_solution_def by force
qed

definition check_NF_iteration :: "('f, 'v) substL \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) redex_prob check" where
  "check_NF_iteration \<mu> \<equiv> let dec = redex_rps_decision \<mu> in (\<lambda>Q t. dec (map (Pair t) Q))"

lemma check_NF_iteration:
  "isOK (check_NF_iteration \<mu> Q t) = (\<forall>i. t \<cdot> mk_subst \<mu> ^^ i \<in> NF_terms (set Q))"
  (is "?l = ?r") 
proof -
  have id: "?r = (\<not>(\<exists>q \<sigma> C i. q \<in> set Q \<and> t \<cdot> mk_subst \<mu> ^^ i = C\<langle>q \<cdot> \<sigma>\<rangle>))" 
    unfolding NF_ctxt_subst by auto
  show ?thesis unfolding id check_NF_iteration_def redex_rps_decision Let_def by force
qed

end
