(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Matrix_Poly
imports 
  Linear_Polynomial 
  Matrix Matrix_Complexity
  SN_Order_Carrier
begin

context matrix_carrier 
begin

lemma mat_linear_poly_order_carrier: assumes sd_n: "sd \<le> n" and sd_pos: "sd > 0" shows 
  "linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gt sd) True (mat1 n) (mat_default n) mat_max0 (\<lambda> x. True)"
  by (rule linear_poly_order_carrier.intro, rule linear_poly_carrier.intro,
    rule mat0, 
    rule mat_plus, simp+,
    rule mat_plus_comm,
    rule mat_plus_assoc, simp+,
    simp only: mat_plus_comm[of "mat0 n n"], rule plus_mat0, simp,
    rule mat_mult, simp+,
    rule mat_mult_assoc_n[symmetric], simp+,
    rule mat0_mult_left, simp, 
    rule mat0_mult_right, simp,
    rule mat_mult_plus_distrib_right, simp+,
    rule mat_mult_plus_distrib_left, simp+,
    rule mat_gt_compat[OF sd_n], simp+,
    rule mat_gt_compat2[OF sd_n], simp+,
    rule mat_plus_left_mono, simp+,
    rule mat_mult_left_mono, simp+,
    rule mat_mult_right_mono, simp+,
    rule mat_ge_refl, 
    rule mat_ge_trans, simp+,
    rule mat_gt_trans[OF sd_n], simp+,
    rule mat_plus_gt_left_mono[OF sd_n _ mat_ge_refl], simp+,
    rule linear_poly_order_carrier_axioms.intro,
    simp add:  mat_sn[OF sd_n],
    rule mat_max0_id_pos, simp+,
    rule mat_max0_pos, simp+,
    rule mat_max0_mono, simp+,
    rule mat_max0_x, simp+,
    rule mat_max0, simp,
    rule mat1,
    rule mat1,
    rule mat1_mult_left, simp,
    rule mat1_mult_right, simp,
    rule mat1_ge_mat0,
    rule mat_default_ge_mat0,
    simp_all)
end

  


context mono_matrix_carrier
begin      
lemma mat_mono_linear_poly_order_carrier: assumes sd_n: "sd \<le> n" and sd_pos: "sd > 0" shows 
  "mono_linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gt sd) True (mat1 n) (mat_default n) mat_max0 (\<lambda> x. True) (mat_mono sd) mat_max"
  by (rule mono_linear_poly_order_carrier.intro, rule mat_linear_poly_order_carrier[OF sd_n sd_pos],
    rule mono_linear_poly_order_carrier_axioms.intro,
    simp,
    rule mat_mono[OF sd_n], 
    (simp_all)[6],
    rule mat_default_gt_mat0[OF sd_pos sd_n], 
    simp, 
    (rule mat_max, simp+)[1], 
    (rule mat_max_ge(1), simp+)[1],
    (rule mat_max_ge(2), simp+)[1])

lemma mat_complexity_linear_poly_order_carrier: assumes sd_n: "sd \<le> n" and sd_pos: "sd > 0" shows 
  "complexity_linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge (mat_gt sd) True (mat1 n) (mat_default n) mat_max0 (\<lambda> x. True) (mat_mono sd) mat_max (\<lambda> a. bound (linear_norm a))"
proof -
  from mat_mono_linear_poly_order_carrier[OF sd_n sd_pos]
  interpret mono_linear_poly_order_carrier "mat n n" "mat0 n n" mat_plus "mat_mult n" mat_ge "mat_gt sd" True "mat1 n" "mat_default n" mat_max0 "\<lambda> x. True" "mat_mono sd" mat_max .
  show ?thesis
  proof(unfold_locales, unfold simp_thms)
    show "bound (linear_norm (mat0 n n)) = 0"
      unfolding linear_norm_0 by (rule bound_zero)
  next    
    fix x y :: "'a mat"    
    assume ge: "mat_ge x y" and x: "mat n n x" and y: "mat n n y"
    show "bound (linear_norm x) \<ge> bound (linear_norm y)"
      by (rule bound_mono, rule linear_norm_ge[OF x y ge])
  next
    fix x y :: "'a mat"    
    assume x: "mat n n x" and y: "mat n n y"
    show "bound (linear_norm (mat_plus x y)) \<le> bound (linear_norm x) + bound (linear_norm y)" unfolding linear_norm_plus[OF x y]
      by (rule bound_plus)
  next
    fix m :: "'a mat"
    let ?m = "mat n n"
    let ?R = "{(x,y). ?m x \<and> ?m y \<and> mat_ge y (mat0 n n) \<and> mat_gt sd x y}"
    let ?f = "\<lambda> m. linear_norm m"
    let ?fR = "{(x,y). y \<succeq> 0 \<and> x \<succ> y}"
    {
      fix x y
      assume "(x,y) \<in> ?R"
      hence x: "?m x" and y: "?m y" and ge: "mat_ge y (mat0 n n)" and gt: "mat_gt sd x y" by auto
      from linear_norm_gt[OF x y gt] linear_norm_ge[OF y mat0 ge, unfolded linear_norm_0]
      have "(?f x, ?f y) \<in> ?fR" by simp
    } note step = this
    show "deriv_bound ?R m (bound (linear_norm m))"
      by (rule deriv_bound_image[of _ "linear_norm", OF bound step])
  qed
qed

lemma linear_poly_order_carrier: "linear_poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) ge gt True 1 d max0 (\<lambda> x. True)"
proof(unfold_locales, simp_all add: field_simps SN max0_x max0_mono max0_pos max0_id_pos plus_left_mono times_left_mono times_right_mono one_ge_zero ge_refl default_ge_zero)
  fix s t u
  assume "s \<succeq> t" and "t \<succ> u"
  thus "s \<succ> u" by (rule compat)
next
  fix s t u
  assume "s \<succ> t" and "t \<succeq> u"
  thus "s \<succ> u" by (rule compat2)
next
  fix s t u
  assume "s \<succeq> t" and "t \<succeq> u"
  thus "s \<succeq> u" by (rule ge_trans)
next
  fix x y z u :: 'a
  assume gt: "x \<succ> y" 
  thus "x + z \<succ> y + z"
    by (rule plus_gt_left_mono)
next
  fix x y z :: 'a
  assume gt: "x \<succ> y" "y \<succ> z"
  thus "x \<succ> z"
    by (rule gt_trans)
qed
end

lemma (in mono_matrix_carrier) mono_linear_poly_order_carrier: "mono_linear_poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) ge gt True 1 d max0 (\<lambda> x. True) mono max"
proof -
  interpret linear_poly_order_carrier "\<lambda> x. True" 0 "op +" "op *" ge gt True 1 d max0 "\<lambda> x. True"
    by (rule linear_poly_order_carrier)
  show ?thesis
    by (unfold_locales, auto simp: mono default_gt_zero)
qed

lemma (in mono_matrix_carrier) complexity_linear_poly_order_carrier: fixes n_of :: "'a \<Rightarrow> nat"
  shows "complexity_linear_poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) ge gt True 1 d max0 (\<lambda> x. True) mono max bound"
proof -
  interpret mono_linear_poly_order_carrier "\<lambda> x. True" 0 "op +" "op *" ge gt True 1 d max0 "\<lambda> x. True" mono max
    by (rule mono_linear_poly_order_carrier)
  show ?thesis
    by (unfold_locales, unfold simp_thms, rule bound, rule bound_plus, rule bound_zero, rule bound_mono)
qed

interpretation int_complexity: mono_matrix_carrier "op > :: int \<Rightarrow> int \<Rightarrow> bool" 1 nat int_mono
proof(unfold_locales, auto)
  fix x
  let ?R = "{(x,y). 0 \<le> (y :: int) \<and> y < x}" 
  show "deriv_bound ?R x (nat x)"
    unfolding deriv_bound_def
  proof
    assume "(\<exists> y. (x,y) \<in> ?R ^^ Suc (nat x))"
    then obtain y where xy: "(x,y) \<in> ?R ^^ Suc (nat x)" ..
    from xy have y: "0 \<le> y" by auto
    obtain n where n: "n = Suc (nat x)" by auto
    from xy[unfolded n[symmetric]]
    have "x \<ge> y + int n"
    proof (induct n arbitrary: x y)
      case 0 thus ?case by auto
    next
      case (Suc n)
      from Suc(2) obtain z where xz: "(x,z) \<in> ?R ^^ n" and zy: "(z,y) \<in> ?R"
        by auto
      from Suc(1)[OF xz] have le: "z + int n \<le> x" .
      from zy have le2: "y + 1 \<le> z" by simp
      with le show ?case by auto
    qed
    with y have nx: "int n \<le> x" by simp
    from nx have x0: "x \<ge> 0" by simp
    with nx n
    show False by simp
  qed      
qed (insert int_SN.mono, auto)

lemma rat_complexity: assumes d0: "d > 0" and d1: "d \<le> 1" and drev: "\<exists> n. d = 1 / of_nat n"
  shows "mono_matrix_carrier (rat_gt d) 1 (\<lambda> x. nat (ceiling (x / d))) rat_mono"
proof -
  let ?gt = "rat_gt d"
  let ?nat = "\<lambda> x. nat (ceiling (x / d))"
  from rat_interpretation[OF d0 d1]
  interpret SN_strict_mono_ordered_semiring_1 1 ?gt rat_mono .
  have id: "(op \<succeq> :: rat \<Rightarrow> rat \<Rightarrow> bool) = (\<lambda> x y. x \<ge> y)" using ge_rat.simps by (intro ext, simp)
  show ?thesis
  proof(unfold_locales, unfold id)
    show "?nat 0 = 0" using d0 by auto
  next
    fix x y :: rat
    assume xy: "x \<ge> y"
    show "?nat x \<ge> ?nat y" 
      by (rule nat_mono, rule ceiling_mono, insert xy d0, auto simp: field_simps)
  next
    fix x y :: rat
    have "ceiling ((x + y) / d) = ceiling (x / d + y / d)" using d0 by (simp add: field_simps)
    also have "... \<le> ceiling (x / d) + ceiling (y / d)" by (rule ceiling_add_le)
    finally show "?nat (x + y) \<le> ?nat x + ?nat y" by auto
  next
    fix x :: rat
    let ?R = "{(x,y). 0 \<le> y \<and> ?gt x y}"
    show "deriv_bound ?R x (?nat x)" unfolding deriv_bound_def
    proof
      assume "(\<exists> y. (x,y) \<in> ?R ^^ Suc (?nat x))"
      then obtain y where xy: "(x,y) \<in> ?R ^^ Suc (?nat x)" ..
      from xy have y: "0 \<le> y" by auto
      obtain n where n: "n = Suc (?nat x)" by auto
      from xy[unfolded n[symmetric]]
      have "x \<ge> y + d * of_nat n"
      proof (induct n arbitrary: x y)
        case 0 thus ?case by auto
      next
        case (Suc n)
        from Suc(2) obtain z where xz: "(x,z) \<in> ?R ^^ n" and zy: "(z,y) \<in> ?R"
          by auto
        from Suc(1)[OF xz] have le: "z + d * of_nat n \<le> x" .
        from zy[unfolded rat_gt_def] have le2: "y + d \<le> z" by simp
        with le show ?case by (auto simp: field_simps)
      qed
      with y have nx: "d * of_nat n \<le> x" by simp
      from d0 have d: "0 \<le> d" by auto
      have "0 \<le> d * of_nat n" by (rule mult_nonneg_nonneg, insert d, auto)
      with nx have x0: "x \<ge> 0" by auto
      have xd0: "0 \<le> x / d"
        by (rule divide_nonneg_pos[OF x0 d0])
      from nx[unfolded n]
      have "d + d * of_nat (?nat x) \<le> x" by (simp add: field_simps)
      with d0 have "d * of_nat (?nat x) < x" by simp 
      also have "of_nat (?nat x) = of_int (ceiling (x / d))" 
        by (rule of_nat_nat, insert xd0, auto)
      also have "of_int (ceiling (x / d)) \<ge> x / d" using ceiling_correct[of "x / d"] by simp
      finally show False using d0 by (auto simp: field_simps)
    qed
  next
    have "1 \<le> nat 1" by simp
    also have "... \<le> ?nat 1"
    proof (rule nat_mono)
      have "1 = ceiling (1 :: rat)" by simp
      also have "... \<le> ceiling (1 / d)"
        by (rule ceiling_mono, insert d1 d0, auto)
      finally show "1 \<le> ceiling (1 / d)" .
    qed
    finally show "1 \<le> ?nat 1" .
  next
    fix x :: rat and n :: nat
    assume x: "0 \<le> x"
    from drev obtain m where d: "d = 1 / of_nat m" by auto
    have "?nat (x + of_nat n) = nat (ceiling (x / d + of_nat n / d))" using d0
      by (simp add: field_simps)
    also have "of_nat n / d = of_int (of_nat (n * m))" unfolding d by (simp add: field_simps of_nat_mult)
    also have "ceiling (x / d + of_int (of_nat (n * m))) = ceiling (x / d) + of_nat (n * m)" unfolding ceiling_add_of_int ..
    also have "nat (ceiling (x / d) + of_nat (n * m)) = ?nat x + nat (int (n * m))" 
    proof (rule int_complexity.bound_plus_of_nat)
      have "x / d \<ge> 0" 
        by (rule divide_nonneg_pos, insert x d0, auto)
      thus "ceiling (x / d) \<succeq> 0" by auto
    qed
    also have "(nat (int (n * m))) = n * m" by simp
    also have "n * m = ?nat (of_nat n)" unfolding d by (simp add: of_nat_mult[symmetric])
    finally
    show "?nat (x + of_nat n) = ?nat x + ?nat (of_nat n)" .
  qed (insert mono, auto)
qed

locale both_mono_matrix_carrier = SN_both_mono_ordered_semiring_1 d gt
  for gt :: "'a :: {show,ordered_semiring_1} \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<succ>" 50) and d :: 'a
begin


lemma mat_linear_poly_order_carrier: assumes n_pos: "n > 0" shows 
  "linear_poly_order_carrier (mat n n) (mat0 n n) mat_plus (mat_mult n) mat_ge mat_gt_arc False (mat1 n) (mat_default n) id (mat_arc_posI arc_pos)"
  by (rule linear_poly_order_carrier.intro, rule linear_poly_carrier.intro, 
    rule mat0, 
    rule mat_plus, simp+,
    rule mat_plus_comm,
    rule mat_plus_assoc, simp+,
    simp only: mat_plus_comm[of "mat0 n n"], rule plus_mat0, simp,
    rule mat_mult, simp+,
    rule mat_mult_assoc_n[symmetric], simp+,
    rule mat0_mult_left, simp, 
    rule mat0_mult_right, simp,
    rule mat_mult_plus_distrib_right, simp+,
    rule mat_mult_plus_distrib_left, simp+,
    rule mat_gt_arc_compat, simp+,
    rule mat_gt_arc_compat2, simp+,
    rule mat_plus_left_mono, simp+,
    rule mat_mult_left_mono, simp+,
    rule mat_mult_right_mono, simp+,
    rule mat_ge_refl, 
    rule mat_ge_trans, simp+, 
    rule mat_gt_arc_trans, simp+,
    rule mat_gt_arc_plus_mono, simp+,
    rule mat_gt_arc_mult_left_mono, simp+,
    rule mat_gt_arc_mult_right_mono, simp+,
    rule mat0_leastI, simp+,
    rule mat0_leastII, simp+,
    rule linear_poly_order_carrier_axioms.intro,
    rule SN_subset[OF mat_gt_arc_SN[OF n_pos]], force,
    simp,
    simp add: mat0_leastIII,
    simp,
    simp add: mat_ge_refl,
    simp,
    rule mat1,
    rule mat1, 
    rule mat1_mult_left, simp,
    rule mat1_mult_right, simp,
    rule mat1_ge_mat0,
    rule mat_default_ge_mat0,
    rule mat_arc_pos_plus[OF n_pos], simp+,
    rule mat_arc_pos_mult[OF n_pos], simp+,
    rule mat_arc_pos_mat_default[OF n_pos],
    simp,
    rule mat_arc_pos_one[OF n_pos],
    rule mat_arc_pos_zero[OF n_pos],
    rule mat_not_all_ge[OF n_pos]) 
end

lemma (in both_mono_matrix_carrier) linear_poly_order_carrier: "linear_poly_order_carrier (\<lambda> x. True) 0 (op +) (op *) ge gt False 1 d id arc_pos"
proof(unfold_locales)
  fix s t u
  assume "s \<succ> t" and "t \<succeq> u"
  thus "s \<succ> u" by (rule compat2)
next
  fix s t u
  assume "s \<succeq> t" and "t \<succeq> u"
  thus "s \<succeq> u" by (rule ge_trans)
next
  fix s t u
  assume "s \<succ> t" and "t \<succ> u"
  thus "s \<succ> u" by (rule gt_trans)
qed (insert not_all_ge, simp_all add: field_simps plus_left_mono times_left_mono times_right_mono zero_leastI SN zero_leastII zero_leastIII one_ge_zero ge_refl arc_pos_default arc_pos_plus arc_pos_mult compat plus_gt_both_mono times_gt_left_mono times_gt_right_mono arc_pos_one arc_pos_zero)
  
 
end
