(*
Copyright 2012 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Orthogonality
imports 
  Critical_Pairs
begin

inductive_set par_rstep :: "('f,'v)trs \<Rightarrow> ('f,'v)trs" for R :: "('f,'v)trs"
  where root_step[intro]: "(s,t) \<in> R \<Longrightarrow> (s \<cdot> \<sigma>,t \<cdot> \<sigma>) \<in> par_rstep R"     
     |  par_step_fun[intro]: "\<lbrakk>\<And> i. i < length ts \<Longrightarrow> (ss ! i,ts ! i) \<in> par_rstep R\<rbrakk> \<Longrightarrow> length ss = length ts 
             \<Longrightarrow> (Fun f ss, Fun f ts) \<in> par_rstep R"
     |  par_step_var[intro]: "(Var x, Var x) \<in> par_rstep R"

lemma par_rstep_refl[intro]: "(t,t) \<in> par_rstep R"
  by (induct t, auto)

lemma all_ctxt_closed_par_rstep: "all_ctxt_closed (par_rstep R)"
  unfolding all_ctxt_closed_def
  by (auto intro: par_step_fun)

lemma ctxt_closed_par_rstep[intro]: "ctxt_closed (par_rstep R)"
proof (rule one_imp_ctxt_closed)
  fix f bef s t aft
  assume st: "(s,t) \<in> par_rstep R"
  let ?ss = "bef @ s # aft"
  let ?ts = "bef @ t # aft"
  show "(Fun f ?ss, Fun f ?ts) \<in> par_rstep R"
  proof (rule par_step_fun)
    fix i
    assume "i < length ?ts"
    show "(?ss ! i, ?ts ! i) \<in> par_rstep R"
      using par_rstep_refl[of "?ts ! i" R] st by (cases "i = length bef", auto simp: nth_append)
  qed simp
qed

lemma subst_closed_par_rstep: "(s,t) \<in> par_rstep R \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> par_rstep R"
proof (induct rule: par_rstep.induct)
  case (root_step s t \<tau>)
  show ?case
    using par_rstep.root_step[OF root_step, of "\<tau> \<circ>s \<sigma>"] by auto
next
  case (par_step_var x)
  show ?case by auto
next
  case (par_step_fun ss ts f)
  show ?case unfolding subst_apply_term.simps
    by (rule par_rstep.par_step_fun, insert par_step_fun(2-3), auto)
qed

lemma R_par_rstep: "R \<subseteq> par_rstep R"
  using root_step[of _ _ R Var] by auto

lemma par_rstep_rsteps: "par_rstep R \<subseteq> (rstep R)^*"
proof
  fix s t 
  assume "(s,t) \<in> par_rstep R"
  thus "(s,t) \<in> (rstep R)^*"
  proof (induct rule: par_rstep.induct)
    case (root_step s t sigma)
    thus ?case by auto
  next
    case (par_step_var x)
    thus ?case by auto
  next 
    case (par_step_fun ts ss f)
    from all_ctxt_closedE[OF rstep_rtrancl_all_ctxt_closed par_step_fun(3) par_step_fun(2)]
    show ?case unfolding par_step_fun(3) by simp
  qed
qed

lemma rstep_par_rstep: "rstep R \<subseteq> par_rstep R"
  by (rule trs_inclusion[OF ctxt_closed_par_rstep subst_closedI R_par_rstep], 
    insert subst_closed_par_rstep, auto)

lemma par_rsteps_rsteps: "(par_rstep R)^* = (rstep R)^*" (is "?P = ?R")
proof
  from rtrancl_mono[OF par_rstep_rsteps[of R]] show "?P \<subseteq> ?R" by simp
  from rtrancl_mono[OF rstep_par_rstep] show "?R \<subseteq> ?P" .
qed

lemma wf_trs_par_rstep: assumes wf: "\<And> l r. (l,r) \<in> R \<Longrightarrow> is_Fun l"
  and step: "(Var x, t) \<in> par_rstep R"
  shows "t = Var x"
  using step
proof (cases rule: par_rstep.cases)
  case (root_step l r \<sigma>)
  from root_step(1) wf[OF root_step(3)] show ?thesis by (cases l, auto)
qed auto

lemma par_rstep_linear_subst: assumes lin: "linear_term l"
  and step: "(l \<cdot> \<sigma>, t) \<in> par_rstep R"
  shows "(\<exists> \<tau>. t = l \<cdot> \<tau> \<and> (\<forall> x \<in> vars_term l. (\<sigma> x, \<tau> x) \<in> par_rstep R) \<or> 
               (\<exists> C l'' l' r'. l = C\<langle>l''\<rangle> \<and> is_Fun l'' \<and> (l',r') \<in> R \<and> (l'' \<cdot> \<sigma> = l' \<cdot> \<tau>) \<and> ((C \<cdot>c \<sigma>) \<langle> r' \<cdot> \<tau> \<rangle>, t) \<in> par_rstep R))"
  using lin step
proof (induction l arbitrary: t)
  case (Var x t)
  let ?tau = "\<lambda> y. t"
  show ?case
    by (rule exI[of _ ?tau], rule disjI1, insert Var(2), auto)
next
  case (Fun f ss)
  let ?ss = "map (\<lambda> s. s \<cdot> \<sigma>) ss"
  let ?R = "par_rstep R"
  from Fun(3)
  show ?case 
  proof (cases rule: par_rstep.cases)
    case (root_step l r \<tau>)
    show ?thesis
    proof (rule exI, rule disjI2, intro exI conjI)
      show "(l,r) \<in> R" by (rule root_step(3))
      show "Fun f ss = \<box>\<langle>Fun f ss\<rangle>" by simp
      show "(Fun f ss) \<cdot> \<sigma> = l \<cdot> \<tau>" by (rule root_step(1))
      show "((\<box> \<cdot>c \<sigma>)\<langle> r \<cdot> \<tau> \<rangle>, t) \<in> ?R" unfolding root_step(2) using par_rstep_refl by simp
    qed simp
  next
    case (par_step_var x)
    thus ?thesis by simp
  next
    case (par_step_fun ts ss1 g)
    hence id: "ss1 = ?ss" "g = f" and len: "length ts = length ss" by auto
    let ?p1 = "\<lambda> \<tau> i. ts ! i = ss ! i \<cdot> \<tau> \<and> (\<forall> x \<in> vars_term (ss ! i). (\<sigma> x, \<tau> x) \<in> ?R)" 
    let ?p2 = "\<lambda> \<tau> i. (\<exists> C l'' l' r'. ss ! i = C\<langle> l''\<rangle> \<and> is_Fun l'' \<and> (l',r') \<in> R \<and> l'' \<cdot> \<sigma> = l' \<cdot> \<tau> \<and> ((C \<cdot>c \<sigma>)\<langle> r' \<cdot> \<tau> \<rangle>, (ts ! i)) \<in> ?R)"
    let ?p = "\<lambda> \<tau> i. ?p1 \<tau> i \<or> ?p2 \<tau> i"
    {
      fix i
      assume i: "i < length ss"
      with par_step_fun(4) id have i2: "i < length ts" by auto
      from par_step_fun(3)[OF i2] have step: "(ss ! i \<cdot> \<sigma>, ts ! i) \<in> par_rstep R" unfolding id nth_map[OF i] .
      from i have mem: "ss ! i \<in> set ss" by auto
      from Fun.prems(1) mem have "linear_term (ss ! i)" by auto
      from Fun.IH[OF mem this step] have "\<exists> \<tau>. ?p \<tau> i" .
    }
    hence "\<forall> i. \<exists> tau. i < length ss \<longrightarrow> ?p tau i" by blast
    from choice[OF this] obtain taus where taus: "\<And> i. i < length ss \<Longrightarrow> ?p (taus i) i" by blast
    show ?thesis
    proof (cases "\<exists> i. i < length ss \<and> ?p2 (taus i) i")
      case True
      then obtain i where i: "i < length ss" and p2: "?p2 (taus i) i" by blast+
      from par_step_fun(2)[unfolded id] have t: "t = Fun f ts" .
      from i have i': "i < length ts" unfolding len .
      from p2 obtain C l'' l' r' where ssi: "ss ! i = C \<langle>l''\<rangle>" and "is_Fun l''" "(l',r') \<in> R" "l'' \<cdot> \<sigma> = l' \<cdot> taus i" 
        and tsi: "((C \<cdot>c \<sigma>) \<langle> r' \<cdot> taus i \<rangle>, ts ! i) \<in> ?R" by blast
      from id_take_nth_drop[OF i, unfolded ssi] obtain bef aft where ss: "ss = bef @ C \<langle> l'' \<rangle> # aft" 
      and bef: "bef = take i ss" 
      and aft: "aft = drop (Suc i) ss" by blast
      let ?C = "More f bef C aft"
      let ?r = "(C \<cdot>c \<sigma>) \<langle> r' \<cdot> taus i \<rangle>"
      let ?sig = "map (\<lambda> s. s \<cdot> \<sigma>)"
      let ?bra = "?sig bef @ ?r # ?sig aft"
      have C: "(?C \<cdot>c \<sigma>) \<langle> r' \<cdot> taus i \<rangle> = Fun f ?bra" by simp
      show ?thesis unfolding ss
      proof (rule exI[of _ "taus i"], rule disjI2, rule exI[of _ ?C], intro exI conjI)
        show "is_Fun l''" by fact
        show "(l',r') \<in> R" by fact
        show "l'' \<cdot> \<sigma> = l' \<cdot> taus i" by fact
        show "((?C \<cdot>c \<sigma>) \<langle> r' \<cdot> taus i \<rangle>, t) \<in> ?R" unfolding C t
        proof  (rule par_rstep.par_step_fun)
          show "length ?bra = length ts" 
            unfolding len unfolding ss by simp
        next
          fix j
          assume j: "j < length ts"
          show "(?bra ! j, ts ! j) \<in> ?R"
          proof (cases "j = i")
            case True
            hence "?bra ! j = ?r" using bef i by (simp add: nth_append)
            thus ?thesis using tsi True by simp
          next
            case False
            from bef i have "min (length ss) i = i" by simp
            hence "?bra ! j = (?sig bef @ (C \<langle> l'' \<rangle> \<cdot> \<sigma>) # ?sig aft) ! j" using False bef i by (simp add: nth_append)
            also have "... = ?sig ss ! j" unfolding ss by simp
            also have "... = ss1 ! j" unfolding id ..
            finally show ?thesis
            using par_step_fun(3)[OF j] by auto
          qed
        qed
      qed simp
    next
      case False
      with taus have taus: "\<And> i. i < length ss \<Longrightarrow> ?p1 (taus i) i" by blast
      from Fun(2) have "is_partition (map vars_term ss)" by simp
      from subst_merge[OF this, of taus] obtain \<tau> where tau: "\<And> i x. i < length ss \<Longrightarrow> x \<in> vars_term (ss ! i) \<Longrightarrow> \<tau> x = taus i x" by auto
      let ?tau = \<tau>
      {
        fix i
        assume i: "i < length ss"
        hence mem: "ss ! i \<in> set ss" by auto
        from taus[OF i] have p1: "?p1 (taus i) i" .
        have id: "ss ! i \<cdot> (taus i) = ss ! i \<cdot> \<tau>"
          by (rule term_subst_eq, rule tau[OF i, symmetric])
        have "?p1 ?tau i"
        proof (rule conjI[OF _ ballI])
          fix x
          assume x: "x \<in> vars_term (ss ! i)"
          with p1 have step: "(\<sigma> x, taus i x) \<in> par_rstep R" by auto
          with tau[OF i x]
          show "(\<sigma> x, ?tau x) \<in> par_rstep R" by simp
        qed (insert p1[unfolded id], auto)
      } note p1 = this
      have p1: "\<And> i. i < length ss \<Longrightarrow> ?p1 \<tau> i" by (rule p1)
      let ?ss = "map (\<lambda> s. s \<cdot> \<tau>) ss"
      show ?thesis unfolding par_step_fun(2) id
      proof (rule exI[of _ \<tau>], rule disjI1, rule conjI[OF _ ballI])
        have "ts = map (\<lambda> i. ts ! i) [0 ..< (length ts)]" by (rule map_nth[symmetric]) 
        also have "...  = map (\<lambda> i. ?ss ! i) [0 ..< length ?ss]" unfolding len using p1 by auto
        also have "... = ?ss" by (rule map_nth)
        finally have ts: "ts = ?ss" .
        show "Fun f ts = Fun f ss \<cdot> \<tau>" unfolding ts by auto
      next
        fix x
        assume "x \<in> vars_term (Fun f ss)"
        then obtain s where s: "s \<in> set ss" and x: "x \<in> vars_term s" by auto
        from s[unfolded set_conv_nth] obtain i where i: "i < length ss" and s: "s = ss ! i" by auto
        from p1[OF i] x[unfolded s]
        show "(\<sigma> x, \<tau> x) \<in> par_rstep R" by blast
      qed
    qed
  qed
qed

       
lemma weakly_orthogonal_main: fixes R :: "('f,string)trs"
  assumes st1: "(s,t1) \<in> par_rstep R" and st2: "(s,t2) \<in> par_rstep R" and weak_ortho:
  "left_linear_trs R" "\<And> b l r. (b,l,r) \<in> critical_pairs R R \<Longrightarrow> l = r" 
  and wf: "\<And> l r. (l,r) \<in> R \<Longrightarrow> is_Fun l"
  shows "\<exists> u. (t1,u) \<in> par_rstep R \<and> (t2,u) \<in> par_rstep R"
proof -
  let ?R = "par_rstep R"
  let ?CP = "critical_pairs R R"
  {
    fix ls ts \<sigma> f r
    assume below: "\<And> i. i < length ls \<Longrightarrow> ((map (\<lambda> l. l \<cdot> \<sigma>) ls) ! i, ts ! i) \<in> ?R"
    and rule: "(Fun f ls, r) \<in> R"
    and len: "length ts = length ls"
    let ?ls = "map (\<lambda> l. l \<cdot> \<sigma>) ls"
    from weak_ortho(1) rule have lin: "linear_term (Fun f ls)" unfolding left_linear_trs_def by auto
    let ?p1 = "\<lambda> \<tau> i. ts ! i = ls ! i \<cdot> \<tau> \<and> (\<forall> x \<in> vars_term (ls ! i). (\<sigma> x, \<tau> x) \<in> par_rstep R)"
    let ?p2 = "\<lambda> \<tau> i. (\<exists> C l'' l' r'. ls ! i = C\<langle>l''\<rangle> \<and> is_Fun l'' \<and> (l',r') \<in> R \<and> (l'' \<cdot> \<sigma> = l' \<cdot> \<tau>) \<and> ((C \<cdot>c \<sigma>) \<langle> r' \<cdot> \<tau> \<rangle>, ts ! i) \<in> par_rstep R)"
    {
      fix i
      assume i: "i < length ls"
      hence i2: "i < length ts" using len by simp
      from below[OF i] have step: "(ls ! i \<cdot> \<sigma>, ts ! i) \<in> ?R" using i by auto
      from i have mem: "ls ! i \<in> set ls" by auto
      from lin i have lin: "linear_term (ls ! i)" by auto
      from par_rstep_linear_subst[OF lin step] have "\<exists> \<tau>. ?p1 \<tau> i \<or> ?p2 \<tau> i" .
    } note p12 = this
    have "\<exists> u. (r \<cdot> \<sigma>, u) \<in> ?R \<and> (Fun f ts, u) \<in> ?R"
    proof (cases "\<exists> i \<tau>. i < length ls \<and> ?p2 \<tau> i")
      case True
      then obtain i \<tau> where i: "i < length ls" and p2: "?p2 \<tau> i" by blast
      from p2 obtain C l'' l' r' where lsi: "ls ! i = C \<langle> l'' \<rangle>" and l'': "is_Fun (l'')" and lr': "(l',r') \<in> R"
        and unif: "l'' \<cdot> \<sigma> = l' \<cdot> \<tau>" and tsi: "((C \<cdot>c \<sigma>) \<langle> r' \<cdot> \<tau> \<rangle>, ts ! i) \<in> ?R"  by blast
      from id_take_nth_drop[OF i] obtain bef aft where ls: "ls = bef @ C \<langle> l'' \<rangle> # aft" and bef: "bef = take i ls" unfolding lsi by auto
      from i bef have bef: "length bef = i" by auto
      let ?C = "More f bef C aft"
      from bef have hp: "hole_pos ?C = i <# hole_pos C" by simp
      have fls: "Fun f ls = ?C \<langle> l'' \<rangle>" unfolding ls by simp
      from mgu_var_disjoint_string_complete[OF unif] obtain \<mu>1 \<mu>2 \<delta> where 
        mgu: "mgu_var_disjoint_string l'' l' = Some (\<mu>1, \<mu>2)" and id: "l'' \<cdot> \<mu>1 = l' \<cdot> \<mu>2" 
        and sigma: "\<sigma> = \<mu>1 \<circ>s \<delta>" and tau: "\<tau> = \<mu>2 \<circ>s \<delta>" by blast
      let ?sig = "map (\<lambda> s. s \<cdot> \<sigma>)"
      let ?r = "(C \<cdot>c \<sigma>) \<langle> r' \<cdot> \<tau>\<rangle>"
      let ?bra = "?sig bef @ ?r # ?sig aft"
      from weak_ortho(2)[OF critical_pairsI[OF rule lr' fls l'' mgu refl refl refl]]      
      have id: "r \<cdot> \<sigma> = (?C \<cdot>c \<sigma>) \<langle>r' \<cdot> \<tau> \<rangle>" unfolding sigma tau by simp      
      also have "... = Fun f ?bra" by simp
      also have "(..., Fun f ts) \<in> ?R"
      proof (rule par_rstep.par_step_fun)
        show "length ?bra = length ts" unfolding len unfolding ls by simp
      next
        fix j
        assume j: "j < length ts"
        show "(?bra ! j, ts ! j) \<in> ?R"
        proof (cases "j = i")
          case True
          hence "?bra ! j = ?r" using bef i by (simp add: nth_append)
          thus ?thesis using tsi True by simp
        next
          case False
          hence "?bra ! j = (?sig bef @ (C \<langle> l'' \<rangle> \<cdot> \<sigma>) # ?sig aft) ! j" using False bef i by (simp add: nth_append)
          also have "... = ?sig ls ! j" unfolding ls by simp
          finally show ?thesis
            using below[OF j[unfolded len]] by auto
        qed
      qed
      finally have step: "(r \<cdot> \<sigma>, Fun f ts) \<in> ?R" .
      show "\<exists> u. (r \<cdot> \<sigma>, u) \<in> ?R \<and> (Fun f ts, u) \<in> ?R" 
        by (rule exI, rule conjI[OF step par_rstep_refl])
    next
      case False
      with p12
      have "\<forall> i. (\<exists> \<tau>. i < length ls \<longrightarrow> ?p1 \<tau> i)" by blast
      from choice[OF this] obtain tau where tau: "\<And> i. i < length ls \<Longrightarrow> ?p1 (tau i) i" by blast
      from lin have "is_partition (map vars_term ls)" by auto
      from subst_merge[OF this, of tau] obtain \<tau> where \<tau>: "\<And> i x. i < length ls \<Longrightarrow> x \<in> vars_term (ls ! i) \<Longrightarrow> \<tau> x = tau i x"
        by blast
      obtain \<delta> where delta: "\<delta> = (\<lambda> x. if x \<in> vars_term (Fun f ls) then \<tau> x else \<sigma> x)" by auto
      {
        fix i
        assume i: "i < length ls"
        from tau[OF i] have p: "?p1 (tau i) i" .
        have id1: "ls ! i \<cdot> tau i = ls ! i \<cdot> \<tau>"
          by (rule term_subst_eq[OF \<tau>[OF i, symmetric]])
        have id2: "... = ls ! i \<cdot> \<delta>"
          by (rule term_subst_eq, unfold delta, insert i, auto)
        have p: "?p1 \<delta> i" using p[unfolded id] using \<tau>[OF i] unfolding id1 id2 using id2 unfolding delta by auto
      } note delt = this
      have r_delt: "(r \<cdot> \<sigma>, r \<cdot> \<delta>) \<in> ?R"
      proof (rule all_ctxt_closed_subst_step[OF all_ctxt_closed_par_rstep])
        fix x
        assume x: "x \<in> vars_term r"
        show "(\<sigma> x, \<delta> x) \<in> ?R" 
        proof (cases "x \<in> vars_term (Fun f ls)")
          case True
          then obtain l where l: "l \<in> set ls" and x: "x \<in> vars_term l" by auto
          from l[unfolded set_conv_nth] obtain i where i: "i < length ls" and l: "l = ls ! i" by auto
          from delt[OF i] x l show ?thesis by auto
        next
          case False
          hence "\<delta> x = \<sigma> x" unfolding delta by auto
          thus ?thesis by auto
        qed
      qed
      {
        let ?ls = "map (\<lambda> l. l \<cdot> \<delta>) ls"
        have "ts = map (\<lambda> i. ts ! i) [0 ..< length ts]" by (rule map_nth[symmetric])
        also have "... = map (\<lambda> i. ts ! i) [0 ..< length ls]" unfolding len by simp
        also have "... = map (\<lambda> i. ?ls ! i) [0 ..< length ?ls]"
          by (rule nth_map_conv, insert delt[THEN conjunct1], auto)
        also have "... = ?ls"
          by (rule map_nth)
        finally have "Fun f ts = Fun f ls \<cdot> \<delta>" by simp
      } note id = this
      have l_delt: "(Fun f ts, r \<cdot> \<delta>) \<in> ?R" unfolding id
        by (rule par_rstep.root_step[OF rule])
      show "\<exists> u. (r \<cdot> \<sigma>, u) \<in> ?R \<and> (Fun f ts, u) \<in> ?R"
        by (intro exI conjI, rule r_delt, rule l_delt)
    qed
  } note root_arg = this
  from st1 st2 show ?thesis
  proof (induct arbitrary: t2 rule: par_rstep.induct)
    case (par_step_var x t2)
    have t2: "t2 = Var x" 
      by (rule wf_trs_par_rstep[OF wf par_step_var])
    show "\<exists> u. (Var x,u) \<in> ?R \<and> (t2, u) \<in> ?R" unfolding t2
      by (intro conjI exI par_rstep.par_step_var, auto)
  next
    case (par_step_fun ts1 ss f t2)
    note IH = this
    show ?case using IH(4)
    proof (cases rule: par_rstep.cases)
      case (par_step_fun ts2)
      from IH(3) par_step_fun(3) have len: "length ts2 = length ts1" by simp
      {
        fix i
        assume i: "i < length ts1"
        hence i2: "i < length ts2" using len by simp
        from par_step_fun(2)[OF i2] have step2: "(ss ! i, ts2 ! i) \<in> ?R" .
        from IH(2)[OF i step2] have "\<exists> u. (ts1 ! i, u) \<in> ?R \<and> (ts2 ! i, u) \<in> ?R" .
      }
      hence "\<forall> i. \<exists> u. (i < length ts1 \<longrightarrow> (ts1 ! i, u) \<in> ?R \<and> (ts2 ! i, u) \<in> ?R)" by blast
      from choice[OF this] obtain us where join: "\<And> i. i < length ts1 \<Longrightarrow> (ts1 ! i, us i) \<in> ?R \<and> (ts2 ! i, us i) \<in> ?R" by blast
      let ?us = "map us [0 ..< length ts1]"
      {
        fix i
        assume i: "i < length ts1"
        from join[OF this] i have " (ts1 ! i, ?us ! i) \<in> ?R" "(ts2 ! i, ?us ! i) \<in> ?R" by auto 
      } note join = this
      let ?u = "Fun f ?us"
      have step1: "(Fun f ts1, ?u) \<in> ?R"
        by (rule par_rstep.par_step_fun[OF join(1)], auto)
      have step2: "(Fun f ts2, ?u) \<in> ?R"
        by (rule par_rstep.par_step_fun[OF join(2)], insert len, auto)
      show ?thesis unfolding par_step_fun(1) using step1 step2 by blast
    next
      case (root_step l r \<sigma>)
      from wf[OF root_step(3)] root_step(1) obtain ls where l: "l = Fun f ls"
        by auto
      from root_step(1) l have ss: "ss = map (\<lambda> l. l \<cdot> \<sigma>) ls" (is "_ = ?ls") by simp
      from root_step(3) l have rule: "(Fun f ls, r) \<in> R" by simp
      from root_step(2) have t2: "t2 = r \<cdot> \<sigma>" .
      from par_step_fun(3) ss have len: "length ts1 = length ls" by simp
      from root_arg[OF par_step_fun(1)[unfolded ss len] rule len]
      show ?thesis unfolding t2 by blast
    qed
  next
    case (root_step l r \<sigma>)
    note IH = this
    from wf[OF IH(1)] IH(1) obtain f ls where l: "l = Fun f ls" and rule: "(Fun f ls,r) \<in> R" 
      by (cases l, auto)
    from IH(2)[unfolded l] show ?case
    proof (cases rule: par_rstep.cases)
      case (par_step_var x)
      thus ?thesis by simp
    next
      case (root_step l' r' \<tau>)
      hence t2: "t2 = r' \<cdot> \<tau>" by auto
      have id: "Fun f ls = \<box>\<langle>Fun f ls\<rangle>" by simp
      from mgu_var_disjoint_string_complete[OF root_step(1)] obtain mu1 mu2 delta where 
        mgu: "mgu_var_disjoint_string (Fun f ls) l' = Some (mu1, mu2)" and sigma: "\<sigma> = mu1 \<circ>s delta"
        and tau: "\<tau> = mu2 \<circ>s delta" by auto
      from weak_ortho(2)[OF critical_pairsI[OF rule root_step(3) id _ mgu refl refl]]
      have "r \<cdot> mu1 = r' \<cdot> mu2" by simp
      hence id: "r \<cdot> \<sigma> = r' \<cdot> \<tau>" unfolding sigma tau by simp
      show ?thesis unfolding t2 id by auto
    next
      case (par_step_fun ts ls' g)
      hence ls': "ls' = map (\<lambda> l. l \<cdot> \<sigma>) ls" and g: "g = f" and len: "length ts = length ls" by auto
      note par_step_fun = par_step_fun[unfolded ls' g len]
      from root_arg[OF par_step_fun(3) rule len]
      show ?thesis unfolding par_step_fun(2) .
    qed
  qed
qed


lemma weakly_orthogonal_par_rstep_CR:
  assumes weak_ortho: "left_linear_trs R" "\<And> b l r. (b,l,r) \<in> critical_pairs R R \<Longrightarrow> l = r" 
   and wf: "\<And> l r. (l,r) \<in> R \<Longrightarrow> is_Fun l"
shows "CR (par_rstep R)"
proof -
  let ?R = "par_rstep R"
  from weakly_orthogonal_main[OF _ _ weak_ortho wf] 
  have diamond: "\<And> s t1 t2. (s,t1) \<in> ?R \<Longrightarrow> (s,t2) \<in> ?R \<Longrightarrow> \<exists> u. (t1,u) \<in> ?R \<and> (t2,u) \<in> ?R" .
  show ?thesis
    by (rule diamond_imp_CR, rule diamond_I, insert diamond, blast)
qed

lemma weakly_orthogonal_rstep_CR:
  assumes weak_ortho: "left_linear_trs R" "\<And> b l r. (b,l,r) \<in> critical_pairs R R \<Longrightarrow> l = r" 
   and wf: "\<And> l r. (l,r) \<in> R \<Longrightarrow> is_Fun l"
shows "CR (rstep R)"
proof -
  from weakly_orthogonal_par_rstep_CR[OF assms] have "CR (par_rstep R)" .
  thus ?thesis unfolding CR_on_def join_def rtrancl_converse par_rsteps_rsteps .
qed

end
