(*
Copyright 2012 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Q_Reduction_Impl
imports
  Q_Reduction
  QDP_Framework_Impl
  Innermost_Impl
begin

definition
  q_reduction_proc ::
    "('dpp, 'f::show, 'v::show) dpp_ops \<Rightarrow>
    ('f,'v)term list \<Rightarrow> 
    'dpp proc"
where
  "q_reduction_proc I Q' dpp \<equiv> 
    let Pb  = dpp_ops.pairs I dpp;
        Rb = dpp_ops.rules I dpp;
        F = map Some (funas_trs_list (Pb @ Rb));
        Q = dpp_ops.Q I dpp;
        rQ = filter (\<lambda> q. root q \<in> set F) Q
 in check_return (do {
  check (dpp_ops.NFQ_subset_NF_rules I dpp) (shows ''innermost rewriting required'');  
  check_wf_trs Rb;
  check_NF_terms_subset Q' rQ <+? (\<lambda> q. shows ''the term '' +@+ shows q +@+ shows '' is missing in Q' '');
  check_NF_terms_subset Q Q' <+? (\<lambda> q. shows ''the term '' +@+ shows q +@+ shows '' is not allowed in Q' '');
  check_allm (\<lambda> lr. check_subseteq (vars_term_list (snd lr)) (vars_term_list (fst lr)) <+? (\<lambda> x. shows ''variable condition in rule violated'')) Pb
  } <+?
  (\<lambda> e. shows ''problem when reducing Q in the DP problem '' +@+ shows_nl +@+ 
  shows_dpp I dpp +@+ shows_nl +@+ shows ''to the set '' +@+ shows_nl 
  +@+ shows_terms' shows shows ''Q':'' Q' +@+ shows_nl +@+ e)) 
     (dpp_ops.mk I (dpp_ops.nfs I dpp) (dpp_ops.minimal I dpp) (dpp_ops.P I dpp) (dpp_ops.Pw I dpp) Q' (dpp_ops.R I dpp) (dpp_ops.Rw I dpp))"

lemma q_reduction_proc: assumes I: "dpp_spec I"
 shows "dpp_spec.sound_proc_impl I (q_reduction_proc I (Q' :: ('f :: show,'v :: show)term list))"
proof -
  from assms interpret dpp_spec I .
  show ?thesis
  proof
    fix d d'    
    assume ok: "q_reduction_proc I Q' d = Inr d'" and fin: "finite_dpp (dpp d')"
    let ?P = "set (P d)"
    let ?Pw = "set (Pw d)"
    let ?Pb = "?P \<union> ?Pw"
    let ?Q = "set (Q d)"
    let ?Q' = "set Q'"
    let ?R = "set (R d)"
    let ?Rw = "set (Rw d)"
    let ?Rb = "?R \<union> ?Rw"
    let ?nfs = "NFS d"
    let ?m = "M d"
    let ?F = "funas_trs ?Pb \<union> funas_trs ?Rb"
    let ?rQ = "{ q. q \<in> ?Q \<and> root q \<in> Some ` ?F}"
    let ?dpp = "\<lambda> q. (?nfs,?m,?P,?Pw,q,?R,?Rw)"
    note ok = ok[unfolded q_reduction_proc_def Let_def dpp_spec_sound, simplified] 
    from ok have inn: "NF_terms ?Q \<subseteq> NF_trs ?Rb" by auto
    from ok have wf: "wf_trs ?Rb" by auto
    from ok have NFQ: "NF_terms ?Q \<subseteq> NF_terms ?Q'" by auto
    from ok have NFrQ: "NF_terms ?Q' \<subseteq> NF_terms ?rQ" by auto
    from ok have vars: "\<And> lr. lr \<in> ?Pb \<Longrightarrow> vars_term (snd lr) \<subseteq> vars_term (fst lr)" by auto
    hence vars: "\<And> l r. (l,r) \<in> ?Pb \<Longrightarrow> vars_term r \<subseteq> vars_term l" by force
    from ok have id: "dpp d = ?dpp ?Q" "dpp d' = ?dpp ?Q'" by auto
    {
      fix s t \<sigma>
      assume Qchain: "min_ichain (?dpp ?Q) s t \<sigma>"
      from min_ichain_reduced_Q[OF inn refl _ _ wf vars Qchain, of ?F]
      have rQchain: "min_ichain (?dpp ?rQ) s t \<sigma>" by auto
      from Qchain have Qchain: "ichain (?dpp ?Q) s t \<sigma>" by auto
      have Q'chain: "ichain (?dpp ?Q') s t \<sigma>"
        by (rule ichain_mono[OF Qchain], insert NFQ, auto)
      {
        fix i
        from rQchain have SN: "?m \<Longrightarrow> SN_on (qrstep ?nfs ?rQ ?Rb) {t i \<cdot> \<sigma> i}" by (auto simp: minimal_cond_def)
        have SN: "?m \<Longrightarrow> SN_on (qrstep ?nfs ?Q' ?Rb) {t i \<cdot> \<sigma> i}"
          by (rule SN_on_mono[OF SN qrstep_mono], insert NFrQ, auto)
      }
      with Q'chain
      have "min_ichain (?dpp ?Q') s t \<sigma>" by (auto simp: minimal_cond_def)
    }
    with fin      
    show "finite_dpp (dpp d)" unfolding id finite_dpp_def by blast
  qed
qed

end